import copy
import math
from .KalmanFilter import KalmanFilter, KalmanFilterXYZ

class Dingwei():
    # 根据距离求坐标位置

    def __init__(self):
        super().__init__()
        self.using_kalman = 1
        self.kalman = {
            '1-3-20': KalmanFilterXYZ(),

        }

    def triangle_pos_xy(self, point_list, has_height_diff=1):
        """
        传入3个点，已知2个点的坐标，和三个点之间的距离，计算第三个点的坐标
        两点坐标为：（x1，y1），（x2，y2），距离为r1，r2
        is_accuracy 代表对应距离（r1，r2）是否是用户输入的

        :param point_list:[
        {
            "client_id":"1-2-1",
            "dist":{"1-2-2":4.00, "1-2-3":3.00 },
            "pos":{"x":0, "y":0},
        },
        {
            "client_id":"1-2-2",
            "dist":{"1-2-1":4.00, "1-2-3":5.00 },
            "pos":{"x":0.00, "y":4.00}
        },
        {
            client_id:"1-2-3",
            "is_accuracy":{"1-2-2":1, "1-2-1":0 }  # 新增数据 ---  1-说明是用户输入的距离，比较准确， 0-自测距离
            "dist":{"1-2-2":5.00, "1-2-1":3.00 }
        }
        ]
        :return:[{"x": x31, "y": y31, 'z': z}, {"x": x32, "y": y32, 'z': z}, ]
        """

        x1 = point_list[0]['pos']['x']
        y1 = point_list[0]['pos']['y']
        if has_height_diff == 1:
            z = point_list[2]['height']
            z1 = point_list[0]['pos']['z']
            z2 = point_list[1]['pos']['z']
        else:
            z = 0
            z1 = 0
            z2 = 0
        x2 = point_list[1]['pos']['x']
        y2 = point_list[1]['pos']['y']
        client_id1 = point_list[0]['client_id']
        client_id2 = point_list[1]['client_id']
        r1 = point_list[2]['dist'][client_id1]
        # r1_is_accuracy = point_list[2]['is_accuracy'][client_id1]  # r1 是否是用户输入的
        r2 = point_list[2]['dist'][client_id2]
        # r2_is_accuracy = point_list[2]['is_accuracy'][client_id2]  # r2 是否是用户输入的
        # TypeError: can't convert complex to float
        try:
            r3 = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)
        except Exception as e:
            print('距离转位置失败: %s %s %s %s', point_list, z, z1, z2)
            point_list[2]['pos'] = [{"x": 0.1, "y": 0.1, "z": z}]
            return point_list

        if ((z - z1) * (z - z1) >= r1 * r1) or ((r2 * r2 <= (z - z2) * (z - z2))):
            point_list[2]['pos'] = [{"x": 0.1, "y": 0.1, "z": z}]
            return point_list

        r1 = math.sqrt(r1 * r1 - (z - z1) * (z - z1))
        r2 = math.sqrt(r2 * r2 - (z - z2) * (z - z2))

        if (r3 > r1 + r2) or (r2 > r3 + r1) or (r1 > r2 + r3):  # 组成不了三角形
            # ZeroDivisionError: float division by zero
            point_list[2]['pos'] = [
                {"x": x1 + (x2 - x1) * (r1 / (r1 + r2)), "y": y1 + (y2 - y1) * (r1 / (r1 + r2)), "z": z}]
            return point_list

        t = (r2 * r2 - r1 * r1 + x1 * x1 + y1 * y1 - x2 * x2 - y2 * y2)
        a = (2 * x1 - 2 * x2)
        b = (2 * y1 - 2 * y2)

        y31 = 0
        y32 = 0
        x31 = 0
        x32 = 0
        if b != 0:
            c = ((t - b * y1) / b)
            a1 = ((a * a + b * b) / b / b)
            b1 = (-(2 * x1 + 2 * a * c / b))
            c1 = x1 * x1 - r1 * r1 + c * c
            b2_4ac = ((b1 * b1 - 4 * a1 * c1))
            if b2_4ac >= 0:

                x31 = (-b1 + b2_4ac ** 0.5) / 2 / a1
                y31 = (t - a * x31) / b
                x32 = (-b1 - b2_4ac ** 0.5) / 2 / a1
                y32 = (t - a * x32) / b
            elif b2_4ac < 0:
                print('triangle_pos_xy位置计算失败: %s %s %s', r1, r2, point_list[0]['dist'][client_id2])
                x31 = 0
                y31 = 0
                x32 = 0
                y32 = 0

            # print(t, a, b, c, a1, b1, c1)
            # print("one:", x31, y31)
            # print("two:", x32, y32)
        elif a != 0:
            c = (t - a * x1) / a
            a1 = (a * a + b * b) / a / a
            b1 = -(2 * y1 + 2 * b * c / a)
            c1 = y1 * y1 - r1 * r1 + c * c
            b2_4ac = ((b1 * b1 - 4 * a1 * c1))
            if b2_4ac < 0:
                print('triangle_pos_xy位置计算失败2: %s %s %s', r1, r2, point_list[0]['dist'][client_id2])

                y31 = 0
                y32 = 0
                x31 = 0
                x32 = 0
            else:
                y31 = (-b1 + (b1 * b1 - 4 * a1 * c1) ** 0.5) / 2 / a1
                x31 = (t - b * y31) / a
                y32 = (-b1 - (b1 * b1 - 4 * a1 * c1) ** 0.5) / 2 / a1
                x32 = (t - b * y32) / a
            # print(t, a, b, c, a1, b1, c1)
            # print("one:", x31, y31)
            # print("two:", x32, y32)

        point_list[2]['pos'] = [{"x": x31, "y": y31, 'z': z}, {"x": x32, "y": y32, 'z': z}, ]
        # print('打印2 ', point_list)
        return point_list

    def calc_NSE_pos(self, jizhan_dist_list):
        N = None
        S = None
        E = None
        for i in jizhan_dist_list:
            if i['direction_point'] == 'N':
                N = i
            elif i['direction_point'] == 'S':
                S = i
            elif i['direction_point'] == 'E':
                E = i
        if (N is None) or (E is None) or (S is None):
            print('NES could not found: %s %s %s', N, E, S)
            return None, None
        N_client_id = N['client_id']
        S_client_id = S['client_id']
        # N S 先确定位置:S定位xy平面的原点
        N_y = N['dist'][S_client_id]
        # S点在原点
        # N点坐标：如果NS高度相同，则N点坐标为NS的距离
        #         如果NS高度不同，则N点坐标为NS的距离- 相应的高度差
        if (N_y * N_y - (N['height'] - S['height']) ** 2) < 0:
            print('NS之间距离太小，小于高度差，定位失败：%s %s %s %s', N_y, N['height'], S['height'], jizhan_dist_list)
            N['pos'] = {'x': 0, 'y': 0.1, 'z': N['height']}
        else:
            N['pos'] = {'x': 0, 'y': (N_y * N_y - (N['height'] - S['height']) ** 2) ** 0.5, 'z': N['height']}
        S['pos'] = {'x': 0, 'y': 0, 'z': S['height']}

        # client_110 = {
        #     'client_id': '1-1-0',
        #     'pos': {'x': - N_y / 5, 'y': 0, 'z': N['height']}
        # }

        # 确定E的位置
        point_list = [N, S, E]
        point_list = self.triangle_pos_xy(point_list)  #

        # 坐标轴 y:S->N x: S->E
        # 所以，E点x坐标应该>0.此处求出2个点，1个x<0, 1个x>0 取x>的一个
        xyz_2 = point_list[2]['pos']
        if len(xyz_2) > 1:
            if xyz_2[0]['x'] > xyz_2[1]['x']:
                point_list[2]['pos'] = xyz_2[0]
            else:
                point_list[2]['pos'] = xyz_2[1]
        elif len(xyz_2) == 1:
            point_list[2]['pos'] = xyz_2[0]
        else:
            print('E pos could not found')
        return jizhan_dist_list

    def Rtls_Cal_Pos2D(self, x1, y1, r1, x2, y2, r2, x3, y3, r3):
        A = [[0, 0], [0, 0]]
        B = [[0, 0], [0, 0]]
        C = [0, 0]
        result = [0, 0]
        A[0][0] = 2.0 * (x1 - x2)
        A[0][1] = 2.0 * (y1 - y2)
        A[1][0] = 2.0 * (x1 - x3)
        A[1][1] = 2.0 * (y1 - y3)

        det = 0
        det = A[0][0] * A[1][1] - A[1][0] * A[0][1]

        if det != 0:

            B[0][0] = A[1][1] / det
            B[0][1] = -A[0][1] / det

            B[1][0] = -A[1][0] / det
            B[1][1] = A[0][0] / det

            C[0] = r2 * r2 - r1 * r1 - x2 * x2 + x1 * x1 - y2 * y2 + y1 * y1
            C[1] = r3 * r3 - r1 * r1 - x3 * x3 + x1 * x1 - y3 * y3 + y1 * y1

            result[0] = B[0][0] * C[0] + B[0][1] * C[1]  # x的值
            result[1] = B[1][0] * C[0] + B[1][1] * C[1]  # y的值
        else:
            print('no pos found error!')
            result[0] = 0
            result[1] = 0

        return result

    def triangle_pos_xy_three_get_one(self, point_list, has_height_diff=1):
        x1 = point_list[0]['pos']['x']
        y1 = point_list[0]['pos']['y']

        x2 = point_list[1]['pos']['x']
        y2 = point_list[1]['pos']['y']

        x3 = point_list[2]['pos']['x']
        y3 = point_list[2]['pos']['y']

        h1 = point_list[0]['pos']['z']
        h2 = point_list[1]['pos']['z']
        h3 = point_list[2]['pos']['z']
        h = point_list[3]['height']

        r1_n = point_list[3]['dist'][point_list[0]['client_id']]
        r2_n = point_list[3]['dist'][point_list[1]['client_id']]
        r3_n = point_list[3]['dist'][point_list[2]['client_id']]

        r1 = (r1_n * r1_n - has_height_diff * (h - h1) * (h - h1)) ** 0.5
        r2 = (r2_n * r2_n - has_height_diff * (h - h2) * (h - h2)) ** 0.5
        r3 = (r3_n * r3_n - has_height_diff * (h - h3) * (h - h3)) ** 0.5

        xy = self.Rtls_Cal_Pos2D(x1, y1, r1, x2, y2, r2, x3, y3, r3)
        return {'x': xy[0], 'y': xy[1], 'z': h}

    def calc_anchor_pos(self, anchor_list):
        """

        :param anchor_list:
        :return:
        """
        # 次基站定位
        anchor_pos_list = self.calc_NSE_pos(anchor_list)
        print('基站NSE位置:', anchor_pos_list)
        other_jizhan_list = []
        NSE_jizhan_list = []
        for i in anchor_pos_list:
            if i['direction_point'] in ['N', 'S', 'E']:
                NSE_jizhan_list.append(i)
            elif i['direction_point'] == None:
                other_jizhan_list.append(i)
        for b in other_jizhan_list:
            sanwei_xyz_list = copy.deepcopy(NSE_jizhan_list)
            sanwei_xyz_list.append(b)
            print('已知三点坐标，通过映射求第四点坐标xyz： %s', sanwei_xyz_list)
            no_four_pos = self.triangle_pos_xy_three_get_one(sanwei_xyz_list)  # 三维 ， 返回第四个点坐标=========================
            # print('da 基站pos', no_four_pos)
            b['pos'] = no_four_pos
        return anchor_pos_list

    # tag=================
    def get_jizhan_list_to_calc_biaoqian(self, jizhan_info=None):
        """

        :param jizhan_info:[{client_id:1-2-0,},...]
        :return: [ ['1-2-0','1-2-1','1-2-2','1-2-3'],  ['1-2-0','1-2-1','1-2-2','1-2-4'],]


        """
        # jizhan_list 用于四个基站定位一个标签
        curr_jizhan_list = [i['client_id'] for i in jizhan_info]
        curr_jizhan_list.sort()
        if len(curr_jizhan_list) <= 4:
            return [curr_jizhan_list]

        jizhan_list = []
        jizhan_list1 = curr_jizhan_list[:3]
        jizhan_list2 = curr_jizhan_list[3:]
        for i in jizhan_list2:
            jizhan_list.append([*jizhan_list1, i])
        # jizhan_list = [
        #     ['1-2-0', '1-2-1', '1-2-2', '1-2-5'],
        #     ['1-2-0', '1-2-1', '1-2-3', '1-2-5'],
        #     ['1-2-0', '1-2-2', '1-2-3', '1-2-5'],
        #     ['1-2-1', '1-2-2', '1-2-3', '1-2-4'],
        # ]
        # jizhan_list = [
        #     ['1-2-0', '1-2-1', '1-2-2', '1-2-3']
        # ]
        return jizhan_list
    def triangle_pos_xyz(self, point_list):
        """
        point_list:[
        {
            "client_id":"1-2-1",
            "dist":{"1-2-2":3.00, "1-2-3":4.00 },
            "pos":{"x":0.00, "y":0.00, "z":0.00}
        },
        {
            "client_id":"1-2-2",
            "dist":{"1-2-1":3.00, "1-2-3":5.00 },
            "pos":{"x":10.00, "y":10.00, "z":0.00}
        },
        {
            client_id:"1-2-3",
            "dist":{1-2-2:5.00, 1-2-1:4.00 },
            "pos":{"x":0.00, "y":10.00, "z":0.00}
        },
        {
            client_id:"1-2-4",
            "dist":{1-2-2:5.00, 1-2-1:4.00 },
            "pos":{"x":10.00, "y":0.00, "z":10.00}
        },
        {
            client_id:"1-2-5",
            "dist":{1-2-1:10*(3**0.5), 1-2-2:10.00, 1-2-3:10*(2**0.5), 1-2-4:10.00}
        }

        ]

        传入5个点，已知4个点的坐标，和5个点之间的距离，计算第5个点的坐标

        :param point_list:[]
        :return:{x:,y:,z:}
        """
        # print(point_list)
        if len(point_list) < 4:
            # logger.error('距离转位置失败 %s', point_list)
            return {'x': 0, 'y': 0, 'z': 0}
        if ('pos' not in point_list[3]) or ('x' not in point_list[3]['pos']):
            # logger.error('距离转位置失败 %s', point_list)
            return {'x': 0, 'y': 0, 'z': 0}

        Point_xyz = [0.0, 0.0, 0.0]
        x1 = point_list[0]['pos']['x']
        y1 = point_list[0]['pos']['y']
        z1 = point_list[0]['pos']['z']
        x2 = point_list[1]['pos']['x']
        y2 = point_list[1]['pos']['y']
        z2 = point_list[1]['pos']['z']
        x3 = point_list[2]['pos']['x']
        y3 = point_list[2]['pos']['y']
        z3 = point_list[2]['pos']['z']
        x4 = point_list[3]['pos']['x']  # x4 = point_list[3]['pos']['x'] KeyError: 'pos'
        y4 = point_list[3]['pos']['y']
        z4 = point_list[3]['pos']['z']
        # print(x1, y1, z1)
        # print(x2, y2, z2)
        # print(x3, y3, z3)
        # print(x4, y4, z4)
        client_id1 = point_list[0]['client_id']
        client_id2 = point_list[1]['client_id']
        client_id3 = point_list[2]['client_id']
        client_id4 = point_list[3]['client_id']
        try:
            r1 = point_list[4]['dist'][client_id1]
            r2 = point_list[4]['dist'][client_id2]
            r3 = point_list[4]['dist'][client_id3]
            r4 = point_list[4]['dist'][client_id4]
        except Exception as e:
            # logger.error('四点求第五点 %s %s', e, point_list)
            return {'x': Point_xyz[0], 'y': Point_xyz[1], 'z': Point_xyz[2]}
        A = [[2 * (x1 - x2), 2 * (y1 - y2), 2 * (z1 - z2)],
             [2 * (x1 - x3), 2 * (y1 - y3), 2 * (z1 - z3)],
             [2 * (x1 - x4), 2 * (y1 - y4), 2 * (z1 - z4)]]
        # print('A', A)
        AT = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        for i in range(3):
            for j in range(3):
                AT[i][j] = A[j][i]
        # print('AT', AT)
        ATA = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        for i in range(3):
            for j in range(3):
                ATA[i][j] = AT[i][0] * A[0][j] + AT[i][1] * A[1][j] + AT[i][2] * A[2][j]
        # print('ATA', ATA)
        det = ATA[0][0] * ATA[1][1] * ATA[2][2] + ATA[0][1] * ATA[1][2] * ATA[2][0] + ATA[0][2] * ATA[1][0] * ATA[2][
            1] - ATA[2][0] * ATA[1][1] * ATA[0][2] - ATA[1][0] * ATA[0][1] * ATA[2][2] - ATA[0][0] * ATA[2][1] * ATA[1][
                  2]
        if det != 0:
            # print('det != 0')
            B = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
            B[0][0] = (ATA[1][1] * ATA[2][2] - ATA[1][2] * ATA[2][1]) / det
            B[0][1] = -(ATA[0][1] * ATA[2][2] - ATA[0][2] * ATA[2][1]) / det
            B[0][2] = (ATA[0][1] * ATA[1][2] - ATA[0][2] * ATA[1][1]) / det

            B[1][0] = -(ATA[1][0] * ATA[2][2] - ATA[1][2] * ATA[2][0]) / det
            B[1][1] = (ATA[0][0] * ATA[2][2] - ATA[0][2] * ATA[2][0]) / det
            B[1][2] = -(ATA[0][0] * ATA[1][2] - ATA[0][2] * ATA[1][0]) / det

            B[2][0] = (ATA[1][0] * ATA[2][1] - ATA[1][1] * ATA[2][0]) / det
            B[2][1] = -(ATA[0][0] * ATA[2][1] - ATA[0][1] * ATA[2][0]) / det
            B[2][2] = (ATA[0][0] * ATA[1][1] - ATA[0][1] * ATA[1][0]) / det

            H = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
            for i in range(3):
                for j in range(3):
                    H[i][j] = B[i][0] * AT[0][j] + B[i][1] * AT[1][j] + B[i][2] * AT[2][j]
            C = [0, 0, 0]
            C[0] = r2 * r2 - r1 * r1 - x2 * x2 + x1 * x1 - y2 * y2 + y1 * y1 - z2 * z2 + z1 * z1
            C[1] = r3 * r3 - r1 * r1 - x3 * x3 + x1 * x1 - y3 * y3 + y1 * y1 - z3 * z3 + z1 * z1
            C[2] = r4 * r4 - r1 * r1 - x4 * x4 + x1 * x1 - y4 * y4 + y1 * y1 - z4 * z4 + z1 * z1
            Point_xyz[0] = H[0][0] * C[0] + H[0][1] * C[1] + H[0][2] * C[2]
            Point_xyz[1] = H[1][0] * C[0] + H[1][1] * C[1] + H[1][2] * C[2]
            Point_xyz[2] = H[2][0] * C[0] + H[2][1] * C[1] + H[2][2] * C[2]
            # Point_xyz[0] = '{:.3f}'.format(Point_xyz[0])
            # Point_xyz[1] = '{:.3f}'.format(Point_xyz[1])
            # Point_xyz[2] = '{:.3f}'.format(Point_xyz[2])
        else:
            # logger.error('四点计算第五点坐标失败: %s', point_list)
            # print('det == 0')
            Point_xyz[0] = 0
            Point_xyz[1] = 0
            Point_xyz[2] = 0
        no_five_pos = {'x': Point_xyz[0], 'y': Point_xyz[1], 'z': Point_xyz[2]}  # 第五个点的坐标
        # print(point_list[4]['client_id'], ' 第五点坐标： ', no_five_pos)
        return no_five_pos

    def dingwei_biaoqian_dist_to_pos(self, jizhan_pos_param, biaoqian_dist, jizhan_list=None, using_kalman=None):
        """

        :param jizhan_pos_param: [
        {'client_id': '1-2-0', 'pos': {'x': 0, 'y': 30, 'z': 0.5}},
        {'client_id': '1-2-1', 'pos': {'x': 0, 'y': 0, 'z': 1.1}},
        {'client_id': '1-2-2', 'pos': {'x': 10, 'y': 0, 'z': 0.5}},
        {'client_id': '1-2-3', 'pos': {'x': 10, 'y': 30, 'z': 1.1}}
    ]
        :param biaoqian_dist: [{'client_id': '1-3-1', 'dist': {'1-2-0': 5 * (10 ** 0.5), '1-2-1': 5 * (10 ** 0.5), '1-2-2': 5 * (10 ** 0.5), '1-2-3': 5 * (10 ** 0.5)，...}, ...}]
        :param jizhan_list: ['1-2-0', '1-2-1', '1-2-2', '1-2-3', ...]  用其中四个基站定位标签---四个基站有多种方案
        :return:[
                    {
                        'client_id':'1-3-20',
                        'pos':{x:, y:, z:}
                    }
                ]
        """

        if jizhan_list is None:
            jizhan_list = self.get_jizhan_list_to_calc_biaoqian(jizhan_pos_param)

        # logger.debug('标签计算位置：%s %s', jizhan_pos_param, biaoqian_dist)
        biaoqian_pos = []
        for t in range(len(biaoqian_dist)):
            biaoqian_pos.append({'x': [], 'y': [], 'z': []})
        for jl in jizhan_list:
            # print('jl', jl)
            c_list_4 = []
            for i in jizhan_pos_param:
                if i['client_id'] in jl:
                    c_list_4.append(i)
            for b in range(len(biaoqian_dist)):
                # 每个标签，进行定位
                sanwei_xyz_5_list = copy.deepcopy(c_list_4)
                sanwei_xyz_5_list.append(biaoqian_dist[b])
                # logger.debug('求第五点坐标：%s', sanwei_xyz_5_list)
                the_five_pos = self.triangle_pos_xyz(sanwei_xyz_5_list)
                biaoqian_pos[b]['x'].append(the_five_pos['x'])
                biaoqian_pos[b]['y'].append(the_five_pos['y'])
                biaoqian_pos[b]['z'].append(the_five_pos['z'])
                # logger.debug('the_five_pos: %s %s', biaoqian_dist[b]['client_id'], the_five_pos)
        # logger.info('位置 多个位置平均前的 ： %s', biaoqian_pos)

        for bp in range(len(biaoqian_pos)):
            sum_x = 0
            sum_y = 0
            sum_z = 0
            for j in range(len(biaoqian_pos[bp]['x'])):
                sum_x += biaoqian_pos[bp]['x'][j]
                sum_y += biaoqian_pos[bp]['y'][j]
                sum_z += biaoqian_pos[bp]['z'][j]
            avg_x = sum_x / len(biaoqian_pos[bp]['x'])
            avg_y = sum_y / len(biaoqian_pos[bp]['y'])
            avg_z = sum_z / len(biaoqian_pos[bp]['z'])
            biaoqian_dist[bp]['pos'] = {'x': avg_x, 'y': avg_y, 'z': avg_z}
        # logger.info('位置 多个位置平均后的： %s', biaoqian_dist)


        for i in biaoqian_dist:
            cid = i['client_id']
            if cid not in self.kalman:
                self.kalman[cid]=KalmanFilterXYZ()

        if ((using_kalman is None) and self.using_kalman) or using_kalman:
            for i in biaoqian_dist:
                cid = i['client_id']

                i['pos']['x'] = self.kalman[cid].set_x(i['pos']['x'])
                i['pos']['y'] = self.kalman[cid].set_y(i['pos']['y'])
                i['pos']['z'] = self.kalman[cid].set_z(i['pos']['z'])
                # logger.debug('kalman pos：cid=%s %s', cid, i['pos'])
        else:  # not using kalmam
            for i in biaoqian_dist:
                cid = i['client_id']
                self.kalman[cid].set_x(i['pos']['x'])
                self.kalman[cid].set_y(i['pos']['y'])
                self.kalman[cid].set_z(i['pos']['z'])

        biaoqian_list = copy.copy(biaoqian_dist)
        # logger.info('位置 卡尔曼滤波后的位置： %s', biaoqian_list)
        return biaoqian_list


__author__ = "songjiangshan"
__copyright__ = "Copyright (C) 2021 songjiangshan \n All Rights Reserved."
__license__ = ""
__version__ = "1.0"
#https://blog.csdn.net/rusi__/article/details/100122350

import logging
import datetime
format_color = '%(asctime)s - %(levelname)s - %(name)s:%(funcName)s:%(lineno)d - %(message)s'
stream_headler = logging.StreamHandler()

file_log_name = datetime.datetime.now().strftime('%Y_%m_%d-%H_%M_%S')

logging.basicConfig(level=logging.DEBUG,
                    format=format_color,
                    datefmt='%Y-%m-%d %H:%M:%S',
                    handlers=[
                        logging.FileHandler('uwb'+file_log_name+'.log',mode='a'),
                        stream_headler
                    ]
                    )


log_colors ={
    logging.DEBUG: "\033[1;34m",  # blue
    logging.INFO: "\033[1;32m",  # green
    logging.WARNING: "\033[1;35m",  # magenta
    logging.ERROR: "\033[1;31m",  # red
    logging.CRITICAL: "\033[1;41m",  # red reverted
}
for i in log_colors:
    logging.addLevelName( i, log_colors[i]+"%s\033[1;0m" % logging.getLevelName(i))


logging.disable(logging.DEBUG)


from .calc_crc import calc_crc,check_crc
import copy
"""
生成和解码modbus报文
"""

def get_read_pkt(start_reg, reg_count, addr=1):
    """
    生成读报文
    :param start_reg:
    :param reg_count:
    :param addr:
    :return:
    """
    # print('read_modbus: ', add2.to_bytes(2, 'big').hex(), reg_count.to_bytes(1, 'big').hex(), id)
    to_w = addr.to_bytes(1, byteorder='big', signed=False) + b'\x03' + start_reg.to_bytes(2, byteorder='big',
                                                                                   signed=False) + reg_count.to_bytes(
        2, byteorder='big', signed=False)

    to_w_crc = calc_crc(to_w)
    to_w = to_w + to_w_crc
    return to_w

def get_write_pkt( start_reg, reg_count, dat_to_write:bytes, addr=1):
    """
    生成写报文
    :param start_reg:
    :param reg_count:
    :param dat_to_write:
    :param addr:
    :return:
    """
    to_r = addr.to_bytes(1, byteorder='big', signed=False) + b'\x10' + start_reg.to_bytes(2, byteorder='big',
                                                                                   signed=False) + reg_count.to_bytes(
        2, byteorder='big', signed=False) + \
           len(dat_to_write).to_bytes(1, byteorder='big', signed=False) + dat_to_write
    to_r_crc = calc_crc(to_r)
    to_r = to_r + to_r_crc
    return to_r



def decode_read_return(payload:bytes, to_read_reg_count):
    ret=payload
    to_read_len = 5 + to_read_reg_count * 2
    if len(payload) == to_read_len:
        if check_crc(ret):
            check_crc_ok = 1
        else:
            check_crc_ok = 0
        ret2 = {'id': int(ret[0]), 'cmd': int(ret[1]), 'byte_cnt': int(ret[2]),
                'data': ret[3:-2], 'crc': ret[-2:], 'raw_data': ret, 'crc_ok': check_crc_ok}
        ret3 = copy.deepcopy(ret2)
        ret3['raw_data'] = ret3['raw_data'].hex()
        ret3['data'] = ret3['data'].hex()
        ret3['crc'] = ret3['crc'].hex()
    else:
        ret2 = {'raw_data': ret, 'crc_ok': 0}
        ret3 = copy.deepcopy(ret2)
        ret3['raw_data'] = ret3['raw_data'].hex()

    return ret2

def decode_write_return(payload:bytes):
    ret=payload
    if len(payload) == 8:
        if check_crc(payload):
            check_crc_ok = 1
        else:
            check_crc_ok = 0
        ret2 = {'id': int(ret[0]), 'cmd': int(ret[1]), 'start': int(ret[2]) * 256 + int(ret[3]),
                'byte_cnt': 2 * (int(ret[4]) * 256 + int(ret[5])),
                'crc': ret[-2:], 'raw_data': ret, 'crc_ok': check_crc_ok}
        ret3 = copy.deepcopy(ret2)
        ret3['raw_data'] = ret3['raw_data'].hex()
        ret3['crc'] = ret3['crc'].hex()
    else:
        ret2 = {'raw_data': ret, 'crc_ok': 0}
        ret3 = copy.deepcopy(ret2)
        ret3['raw_data'] = ret3['raw_data'].hex()

    return ret2

def split_packet(buf: bytes, pkt_length, read_error_stat):
    read_error_cnt = 0
    ret_pkt_process = 0
    ret_byte_process = 0
    pkt_list = []
    read_ok_cnt = 0

    while True:
        if len(buf) < pkt_length:
            # 报文不够长
            read_error_cnt += 1
            if 'less' not in read_error_stat:
                read_error_stat['less'] = 0
            read_error_stat['less'] += 1
            break
        if buf[0] != 1:
            # 报文头不对
            buf = buf[1:]
            read_error_cnt += 1
            ret_byte_process += 1

            if 'start' not in read_error_stat:
                read_error_stat['start'] = 0
            read_error_stat['start'] += 1
            continue
        if buf[1] != 3:
            # 报文头不对 01 03
            buf = buf[1:]
            read_error_cnt += 1
            ret_byte_process += 1

            if 'read' not in read_error_stat:
                read_error_stat['read'] = 0
            read_error_stat['read'] += 1
            continue
        if buf[2] != 0x2A:
            # 报文头不对 01 03 2A
            buf = buf[1:]
            ret_byte_process += 1
            read_error_cnt += 1

            if 'len' not in read_error_stat:
                read_error_stat['len'] = 0
            read_error_stat['len'] += 1
            continue

        if not check_crc(buf[:pkt_length]):
            print('crc error. ', buf[:pkt_length].hex(),
                  calc_crc(buf[:pkt_length - 2]).hex())
            buf = buf[1:]
            read_error_cnt += 1
            ret_byte_process += 1

            if 'crc' not in read_error_stat:
                read_error_stat['crc'] = 0
            read_error_stat['crc'] += 1
            continue

        read_ok_cnt += 1
        ret_pkt_process += 1
        ret_byte_process += pkt_length
        pkt_list.append(buf[0:pkt_length])
        if len(buf) > pkt_length:
            buf = buf[pkt_length:]
        else:
            buf = b''

    return pkt_list, buf, read_error_cnt, ret_pkt_process, ret_byte_process


def test_read_pkt1():
    pass


def test_write_pkt1():
    pass

if __name__=='__main__':
    pass





__author__ = "songjiangshan"
__copyright__ = "Copyright (C) 2021 songjiangshan \n All Rights Reserved."
__license__ = ""
__version__ = "1.0"

import serial
import string
import binascii
import time
from .calc_crc import calc_crc
import serial.tools.list_ports
from .mylog import logging
logger = logging.getLogger(__name__)


def is_anchor0(port_name):
    logger.debug('is_anchor0:%s', port_name)
    time.sleep(0.1)
    try:
        s = serial.Serial(port_name, 115200)  # 基站
    except Exception as e:
        # print('open port error', port_name, e)
        return False

    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        #串口有数据，说明正在通讯？
        d1 = bytes.fromhex('01 10 00 28 00 01 02 00 00')
        d1_crc = calc_crc(d1)
        d1 = d1+d1_crc
        s.write(d1)
        time.sleep(0.5)
        count = s.inWaiting()
        s.read(count)
        count = s.inWaiting()
        s.read(count)



    d1 = bytes.fromhex('01 03 00 03 00 01')
    d1_crc = calc_crc(d1)
    d1 = d1 + d1_crc
    s.write(d1)
    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        n = s.read(s.in_waiting)
        logger.debug('%s', n.hex())
        time.sleep(0.02)
        if n == b"\x01\x03\x02\x00\x029\x85":
            s.close()
            logger.debug('OK')
            return True

    s.close()
    logger.debug('NO DETECTED')
    return False

def is_anchor0_old(port_name):
    logger.debug('is_anchor0:%s', port_name)
    time.sleep(0.1)
    try:
        s = serial.Serial(port_name, 115200)  # 基站
    except Exception as e:
        # print('open port error', port_name, e)
        return False

    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        #串口有数据，说明正在通讯？
        d1 = bytes.fromhex('01 10 00 28 00 01 02 00 00')
        d1_crc = calc_crc(d1)
        d1 = d1+d1_crc
        s.write(d1)
        time.sleep(0.5)
        count = s.inWaiting()
        s.read(count)
        count = s.inWaiting()
        s.read(count)



    d1 = bytes.fromhex('01 03 00 03 00 01')
    d1_crc = calc_crc(d1)
    d1 = d1 + d1_crc
    s.write(d1)
    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        n = s.read(s.in_waiting)
        logger.debug('%s', n.hex())
        time.sleep(0.02)
        if n == b"\x01\x03\x02\x00\x029\x85":
            s.close()
            logger.debug('OK')
            return True

    s.close()
    logger.debug('NO DETECTED')
    return False

def is_print(port_name):
    logger.debug('is_print:%s', port_name)
    time.sleep(0.1)
    try:
        s = serial.Serial(port_name, 256000)  # 基站
    except Exception as e:
        # print('open port error', port_name, e)
        return False

    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        #串口有数据，说明正在通讯？
        d1 = bytes.fromhex('01 10 00 28 00 01 02 00 00')
        d1_crc = calc_crc(d1)
        d1 = d1+d1_crc
        s.write(d1)
        time.sleep(0.5)
        count = s.inWaiting()
        s.read(count)
        count = s.inWaiting()
        s.read(count)



    d1 = bytes.fromhex('01 03 00 03 00 01')
    d1_crc = calc_crc(d1)
    d1 = d1 + d1_crc
    s.write(d1)
    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        n = s.read(s.in_waiting)
        logger.debug('%s', n.hex())
        time.sleep(0.02)
        # 01 03
        # 02 （长度）
        # 00 03 （3为打印工具）
        # F8 45 （crc）
        if n == b"\x01\x03\x02\x00\x03\xF8\x45":
            s.close()
            logger.debug('OK')
            return True

    s.close()
    logger.debug('NO DETECTED')
    return False

def is_alert_light(port_name):
    logger.debug('is_alert_light:%s', port_name)
    time.sleep(0.1)
    try:
        s = serial.Serial(port_name, 9600)  # 基站
    except Exception as e:
        logger.info('open port error %s %s', port_name, e)
        return False
    d1 = bytes.fromhex('21242528')# stop beep &RGB color light
    s.write(d1)
    s.close()
    time.sleep(0.5)

    s = serial.Serial(port_name, 9600)  # 基站

    d1 = bytes.fromhex('21')
    s.write(d1)
    time.sleep(0.1)
    count = s.inWaiting()  # 获取串口缓冲区数据
    if count != 0:
        logger.info('got data count:%d', count)
        time.sleep(0.02)
        n = s.read(s.in_waiting)
        logger.info('got data: %s', n.hex())
        if n == b"!\x00" or n == b'!\x01':
            s.close()
            return True


    logger.info('no got expected data')
    s.close()
    return False


def detect_com_port():
    """
    检测串口：
    :return: ['COM3', 'COM4'] 对应uwb的主基站串口，报警灯串口。
            [None, 'COM4'] ['COM3', None] [None, None] 没有相应设备时，为None
    """
    ret = [None, None, None]
    port_list = list(serial.tools.list_ports.comports())
    if len(port_list) <= 0:
        # print("The Serial port can't find!")
        return ret

    for i in list(port_list):
        port_name = i[0]
        if is_anchor0(port_name):
            ret[0] = port_name
        elif is_alert_light(port_name):
            ret[1] = port_name
        elif is_print(port_name):
            ret[2] = port_name

    logger.info('com port detect result:%s', ret)
    return ret


if __name__ == '__main__':
    print(detect_com_port())





__author__ = "songjiangshan"
__copyright__ = "Copyright (C) 2021 songjiangshan \n All Rights Reserved."
__license__ = ""
__version__ = "1.0"


DEVICE_TYPE_TAG=0 #OLD3
DEVICE_TYPE_ANCHOR=1 #OLD2
DEVICE_TYPE_ANCHORZ=2 #OLD1

def client_id_remove_group(client_id):
    return str(client_id_get_type(client_id)) + '-' + str(client_id_get_no(client_id))


def client_id_get_no(client_id: str):
    """
    输入字符串，返回对应的号
    :param client_id:
    :return: int型数值
    """
    id = int(client_id.split('-')[-1])

    return id


def client_id_get_type(client_id):
    id = int(client_id.split('-')[-2])
    return id


def client_id_get_group(client_id):
    id = int(client_id.split('-')[0])
    return id


def pack_client_id(group_id=0, type_int=DEVICE_TYPE_ANCHOR, no=0):
    ret = str(group_id) + '-' + str(type_int) + '-' + str(no)
    return ret


if __name__ == '__main__':
    assert client_id_get_no('1-2-3') == 3
    assert client_id_get_no('1-2-20') == 20
    assert client_id_get_no('1-12-21') == 21
    print(client_id_get_type('1-12-21'))

    # assert client_id_get_type('1-2-3') == 2
    # assert client_id_get_group('1-2-3') == 1
    # assert client_id_get_group('2-2-3') == 2



import copy
import time
import math
#from numpy import short
from .my_short import my_short as short
from .anchor_locate_algorithm1 import AnchorLocateAlgorithm1
from .tag_locate_algorithm1 import TagLocateAlgorithm1
from .uwb_modbus import UwbModbus,ANCHORZ_NO
from .modbus_pkt import split_packet

from .client_id_utils import client_id_get_no, client_id_get_type, pack_client_id, client_id_remove_group, DEVICE_TYPE_TAG, DEVICE_TYPE_ANCHOR, DEVICE_TYPE_ANCHORZ
from .dist_list_dict_convert import dist_list2dict

from .mylog import logging

logger = logging.getLogger(__name__)


def get_cali_param_all():
    return {
        '1-0': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '1-1': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '1-2': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '1-3': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '1-4': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '1-5': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-1': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-2': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-3': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-4': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-5': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-6': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-20': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
        '0-21': {'0': 1.0, '1': 1.0, '2': 1.0, '3': 1.0, '4': 1.0, '5': 1.0, },
    }


class uwb_zrzn(UwbModbus):

    def __init__(self):
        super(uwb_zrzn, self).__init__()

        self.all_device_list = []
        self.all_device_dict = {}
        self.curr_anchor_to_measure = None  # 当前正在测距的次基站
        self.battery_anchor = [255, 255, 255, 255, 255, 255, 255, 255]
        self.battery_tag = {20: 255, 21: 255, 22: 255, 23: 255, 24: 255, 25: 255, 1: 255, 2: 255, 3: 255, 4: 255,
                            5: 255, 6: 255}
        self.measure_mode = 0  # mode=3,4时，才会接收并处理数据
        self.curr_used_anchor = []
        self.set_cache_data_to_default()
        self.cali_param = get_cali_param_all()

        self.anchor_locate_algorithm = AnchorLocateAlgorithm1()
        self.tag_locate_algorithm = TagLocateAlgorithm1()
        self.group_id = 1  # device group
        self.pos_callback = None


    def set_pos_callback(self, callback_func):
        """
        set the callback function when dist and pos is successfully measured
        :param callback_func: def function_to_call(tag_id:int, dist:dict, pos:dict) pos:{x:1,y:2,z:3}
        :return:
        """
        self.pos_callback = callback_func

    # 检测周围所有设备
    def detect_device(self, tag_no_list=None, anchor_no_list=None):
        """
        tag_no_list: 用户认为存在的标签号 [1,2,3,11]
        anchor_no_list： 用户认为存在的基站 [0,1,2]
        """
        not_exist_anchor_no_list = []  # 不存在的基站
        for i in range(7):
            if i not in anchor_no_list:
                not_exist_anchor_no_list.append(i)
        # 标签测距
        self.set_to_start(tag_no_list, 1, 6, not_exist_anchor_no_list)
        self.locate_loop(tag_no_list, 1, 5)
        self.stop_measure_h()
        # 处理标签距离信息，获取在线设备
        tag_dict = {}
        for i in self.tag_list:
            logger.info('标签测距值(主基站、0,1,2,3...)： %s %s', i, self.tag_list[i]['dist'])
            tag_dict[i] = self.tag_list[i]['dist']
        ret = self._get_device_exist_using_dist(tag_dict)
        # 设置在线设备
        tag_no_list = ret['exist_tag']
        anchor_no_list = ret['exist_anchor']
        self.set_device(tag_no_list, anchor_no_list)
        return tag_no_list, anchor_no_list


    def _get_device_exist_using_dist(self, tag_dict):
        """
        分析哪些设备在线
        :param tag_dict: {20: [1,1,2,3,-1,-1,-1,-1],    -----------1-1-0、1-2-0、1-2-1...
                          21: [1,1,2,3,-1,-1,-1,-1],
                          22: [-1,-1,-1,-1,-1,-1,-1,-1]}
        :return: {'exist_anchor': [0, 1, 2, 3], 'exist_tag': [20, 21]}
        """
        ret = {'exist_anchor': [], 'exist_tag': [], 'is_exist_main_anchor': 1}  # is_exist_main_anchor主基站是否存在：1在
        for i in tag_dict:
            for index, val in enumerate(tag_dict[i]):
                if val != -1:
                    ret['exist_tag'].append(i)
                    if index==0:
                        #主基站不写入，因为必须存在，而不进行定位
                        continue
                    ret['exist_anchor'].append(index)

        ret['exist_tag'] = list(set(ret['exist_tag']))
        ret['exist_anchor'] = list(set(ret['exist_anchor']))
        logger.info('存在设备：%s', ret)
        return ret

    def detect_device_slow(self, tag_no_list=None, anchor_no_list=None):
        """
        detect if device exist
        如果是空列表，则基站会检测：0,1,2,3,4,5,6,7存在不存在，标签会检测：1-10存在不存在
        :param tag_no_list: tag_no will be detected. will detect 1-10 tag is exist or not if None or empty.  如果是空标签会检测：1-10号标签是否存在
        :param anchor_no_list: anchor_no will be detected. will detect 0-7 anchor is exist or not if None or empty. 如果为空，则会检测0-7号基站是否存在
        :return:exist tag_no_list, exist anchor_no_list
        """
        if tag_no_list is None or tag_no_list == []:
            tag_no_list = [i + 1 for i in range(10)]
        if anchor_no_list is None or anchor_no_list == []:
            anchor_no_list = [i for i in range(8)]

        tag_client_list = [pack_client_id(self.group_id, DEVICE_TYPE_TAG, i) for i in tag_no_list]
        anchor_client_list = [pack_client_id(self.group_id, DEVICE_TYPE_ANCHOR, i) for i in anchor_no_list]

        """检测设备是否存在"""
        # 避免上次通讯错误，导致0标签的存在
        self.bulk_mode = 0
        self.convert_tag_to_anchor_h('0-0-0', 3)
        for i in anchor_client_list.copy():
            client_id = i
            if self.is_demo_mode:
                loop_max = 1
            else:
                loop_max = 3
            if self.is_anchor_exist_h(client_id, loop_max):
                pass
            else:
                print('基站' + client_id + '测试通讯失败.去除此设备')
                anchor_client_list.remove(client_id)

        time.sleep(1)
        self.if_tag_exist_h('1-3-1')
        time.sleep(1)

        for i in tag_client_list.copy():
            client_id = i
            if self.is_demo_mode:
                loop_max = 1
            else:
                loop_max = 5
            if self.if_tag_exist_h(client_id, loop_max):
                pass
            else:
                print('标签' + client_id + '测试通讯失败.去除此设备')
                tag_client_list.remove(client_id)

        tag_no_list = [client_id_get_no(i) for i in tag_client_list]
        anchor_no_list = [client_id_get_no(i) for i in anchor_client_list]
        self.set_device(tag_no_list, anchor_no_list)
        return tag_no_list, anchor_no_list

    # 显示当前所有设备及相关测距信息
    def device_pos_list(self):
        """
        get all device current pos
        :return:[{'client_id':'1-2-3', pos:{x:,y:,z:}},...]
        """
        # print('所有使用的设备信息：', self.all_device_list)
        ret = []
        for i in self.all_device_list:
            d = {}
            d['client_id'] = i['client_id']
            d['pos'] = i['pos']
            ret.append(d)
        return ret

    def tag_no_list(self):
        """
        get current tag no list
        :return:[1,2,3,...]
        """
        used_tag_no_list = []
        for i in self.all_device_list:
            tag_type = client_id_get_type(i['client_id'])
            tag_no = client_id_get_no(i['client_id'])
            if tag_type == DEVICE_TYPE_TAG:
                used_tag_no_list.append(tag_no)
        return used_tag_no_list

    def anchor_no_list(self):
        """
        get current anchor no list
        :return: [0,1,2,3,...]
        """
        used_anchor_no_list = []
        for i in self.all_device_list:
            tag_type = client_id_get_type(i['client_id'])
            tag_no = client_id_get_no(i['client_id'])
            if tag_type == DEVICE_TYPE_ANCHOR:
                used_anchor_no_list.append(tag_no)
        return used_anchor_no_list

    def set_device(self, tag_no_list, anchor_no_list):
        """
        set device to use.
        this function will auto called when  you call detect_device.
        如果不需要检测设备，直接指定设备列表也可以 (tag_no_list：标签列表，anchor_no_list：次基站列表，using_master：是否使用主基站测距【1代表使用，0代表不使用】)
        暂不支持通过master进行定位
        :param tag_no_list: tag no in the system. eg:[1,2,3,...]
        :param anchor_no_list: anchor no in the system eg: [0,1,2,3]
        :return:None
        """
        self.all_device_list = []
        for i in tag_no_list:
            tag_client_id = pack_client_id(self.group_id, DEVICE_TYPE_TAG, i)
            self.all_device_list.append({'client_id': tag_client_id, 'dist': [], 'pos': []})
        for i in anchor_no_list:
            anchor_client_id = pack_client_id(self.group_id, DEVICE_TYPE_ANCHOR, i)
            self.all_device_list.append({'client_id': anchor_client_id, 'dist': [], 'pos': []})

        self.all_device_dict = {i['client_id']: i for i in self.all_device_list}

        # 使能设备
        for i in self.anchor_used_dict.copy():
            if i not in anchor_no_list:
                self.anchor_used_dict[i] = 0
            else:
                self.anchor_used_dict[i] = 1
        logger.info('当前启动的基站列表: %s', self.anchor_used_dict)

        return None

    def locate_anchor(self, height_list=(0.5, 1.1, 0.5, 1.1), direction_point=('N', 'S', 'E'), measure_ready_cnt=40):
        """
        对基站坐标进行定位：
        1 测距：各基站到其他基站的距离：
        2 计算位置：根据距离和高度参数，以及顺序位置，建立坐标系
        :param height_list:高度列表，需要和设备数一致
        :param measure_ready_cnt:测距次数，多次测量取均值，用于提高测距精度，进而提高定位精度
        :return:
        """
        # 定位次基站
        anchor_list, tag_info = self._get_anchor_and_tag_info()

        # 次基站之间测距
        anchor_list = self._measure_all_anchor_dist(anchor_list, measure_ready_cnt)
        print('基站之间测距：', anchor_list)
        for i in range(len(anchor_list)):
            if i < len(direction_point):
                anchor_list[i]['direction_point'] = direction_point[i]
            else:
                anchor_list[i]['direction_point'] = None
            if i < len(height_list):
                anchor_list[i]['height'] = height_list[i]
            else:
                anchor_list[i]['height'] = None

        self.anchor_locate_algorithm.set_anchor_info(anchor_list)
        self.anchor_locate_algorithm.calc()
        anchor_pos_list = self.anchor_locate_algorithm.get_anchor_pos()

        # FIXME:此处调用set_anchor_location 进行位置设置-2022.3.23
        self.set_anchor_location(anchor_pos_list)

        return anchor_pos_list

    def set_anchor_location(self, anchor_pos, using_master=0):
        """

        :param anchor_pos:[{client_id:'1-2-0', pos:{x:, y:, z:}} ]
        :param using_master:False/True
        :return:
        """
        # 如果不检测定位基站，或者上次已经检测的定位过，则可以直接设置基站的坐标位置
        anchor_pos_list = anchor_pos
        # 将基站位置同步到内存 self.all_device_list
        for i in range(len(self.all_device_list)):
            for j in anchor_pos_list:
                if self.all_device_list[i]['client_id'] == j['client_id']:
                    self.all_device_list[i]['pos'] = j['pos']
        # 显示基站位置

    # 定位一次 标签
    def start_locate_once(self):
        anchor_client_list, tag_client_list = self._get_anchor_and_tag_info()
        # 标签与次基站之间的距离
        self._measure_all_tag_dist(tag_client_list)

        #FIXME: remove anchorz. as it do not for locating

        ret_tag_dict = {}
        # FIXME: 此处获取各标签位置，传入 定位算法函数，获取返回值-2022.3.23
        # 各标签列表：[{client_id:, dist:{1-2-0:50,1-2-1:30},...},...]
        for i in tag_client_list:
            tag_client_id = i['client_id']
            tag_dist = self.tag_list[client_id_get_no(tag_client_id)]['dist']
            ret = self.tag_locate_algorithm.calc(tag_dist, tag_client_id, anchor_client_list)
            self.all_device_dict[tag_client_id]['pos'] = ret
            ret_tag_dict[tag_client_id] = ret

        # 结束
        return ret_tag_dict

    def measure_distance(self, client_id1, client_id2, ready_cnt=30):
        """
        测量获取距离
        :param client_id1:
        :param client_id2:
        :param ready_cnt: 测距次数。默认测量10次取均值返回
        :return: 两个设备间的距离，单位m
        """
        if client_id_get_type(client_id2) == DEVICE_TYPE_ANCHORZ:
            #swap
            client_id1, client_id2 = client_id2, client_id1
        if (client_id_get_type(client_id1) == DEVICE_TYPE_TAG) and (client_id_get_type(client_id2) == DEVICE_TYPE_TAG):
            #两个都是标签，没法测距，返回错误
            return -1

        if client_id_get_type(client_id1) == DEVICE_TYPE_TAG:
            #swap
            client_id1, client_id2=client_id2, client_id1


        if client_id_get_type(client_id2) == DEVICE_TYPE_ANCHOR:
            # 两个次基站 测距
            anchor_dict=[{'client_id': client_id2, 'dist': {}}]
            if client_id_get_type(client_id1) != DEVICE_TYPE_ANCHORZ:
                anchor_dict.append({'client_id': client_id1, 'dist': {}})
            self._measure_all_anchor_dist(anchor_dict, ready_cnt)
            for i in self.all_device_list:
                if i['client_id'] == client_id2:
                    return i['dist_list'][client_id_get_no(client_id2)]
                elif i['client_id'] == client_id2:
                    return i['dist_list'][client_id_get_no(client_id1)]
        elif client_id_get_type(client_id2) == DEVICE_TYPE_TAG:
            tag_client_id = client_id2
            anchor_client_id = client_id1
            # 标签和基站之间测距
            dist = self._measure_all_tag_dist(tag_client_list=[{'client_id': tag_client_id, 'dist': {}}],
                                              anchor_client_list=[{'client_id': anchor_client_id, 'dist': {}}],ready_cnt=ready_cnt)
            return dist[0][anchor_client_id]

        logging.error('unknown device type: %s %s', client_id1, client_id2)
        return -1




        # 测 次基站和次基站之间的距离 或 次基站和标签之间的距离 或主基站和标签之间的距离，主基站和次基站之间的距离
        # 不能测量：标签和标签之间的距离



        if client_id_get_type(client_id1) == DEVICE_TYPE_TAG or client_id_get_type(client_id2) == DEVICE_TYPE_TAG:
            # 存在 标签
            if client_id_get_type(client_id1) == DEVICE_TYPE_TAG:
                tag_client_id = client_id1
                anchor_client_id = client_id2
            else:
                tag_client_id = client_id2
                anchor_client_id = client_id2

            # 标签和基站之间测距
            dist = self._measure_all_tag_dist(tag_client_list=[{'client_id': tag_client_id, 'dist': {}}],
                                              anchor_client_list=[{'client_id': anchor_client_id, 'dist': {}}])
            return dist[0][anchor_client_id]
            # for i in self.all_device_list:
            #    if i['client_id'] == tag_client_id:
            #        return i['dist'][client_id_get_no(anchor_client_id)]
        else:
            # 两个次基站 测距
            self._measure_all_anchor_dist([{'client_id': client_id1, 'dist': {}}, {'client_id': client_id2, 'dist': {}}])
            for i in self.all_device_list:
                if i['client_id'] == client_id1:
                    return i['dist_list'][client_id_get_no(client_id2)]
                elif i['client_id'] == client_id2:
                    return i['dist_list'][client_id_get_no(client_id1)]

        print('打印get_distance=================', self.all_device_list)

    def _get_anchor_and_tag_info(self):
        """

        :return: [{client_id:, pos:{x:,y:,z:}}], [{client_id:, pos:{x:,y:,z:}}]
        """
        # 从内存self.all_device_list中获取 基站 和 标签信息
        tag_client_list = []
        anchor_client_list = []
        for i in self.all_device_list:
            if client_id_get_type(i['client_id']) == DEVICE_TYPE_ANCHOR:
                anchor_client_list.append(i)
            elif client_id_get_type(i['client_id']) == DEVICE_TYPE_TAG:
                tag_client_list.append(i)
        return anchor_client_list, tag_client_list

    def _measure_all_anchor_dist(self, anchor_client_list, measure_ready_cnt=40):
        """
        基站的测距：
        :param anchor_client_list: [ {client_id:'1-2-3', }]
        :return:
        """
        all_anchor_client_id_list = [i['client_id'] for i in anchor_client_list]
        ret = []
        for i in all_anchor_client_id_list:
            reti = {'client_id': i}
            if client_id_get_no(i)==0:
                # 0 anchor could not change to tag and measure
                continue
            reti['dist'] = self._measure_one_anchor_dist(i, measure_ready_cnt, False)
            ret.append(reti)
        print("基站自动测距结束")
        return ret

    def set_to_start(self, tag_no_list=None, ready_cnt=30, start_cmd=6,except_anchor_list=None):

        if tag_no_list is None:
            tag_no_list = self.tag_no_list()
        self.set_basic_info_h(tag_no_list,except_anchor_list)  # self.uwb.set_cache_data_to_default()内部会调用此函数


        self.tag_list_cache_ready_cnt = ready_cnt


        self.start_measure_h(start_cmd)

    def locate_loop(self, tag_no_list=None, wait_for_finish=0, timeout=10.0):
        """
        测量循环
        :param tag_no_list:
        :param wait_for_finish:
        :param timeout:
        :return:
        """
        # 多次测量取平均值

        if tag_no_list is None:
            tag_no_list = self.tag_no_list()
        cijizhan_no_list = self.anchor_no_list()

        # FIXME: 此处需要次基站和主基站 两种基站的个数
        if  0 not in cijizhan_no_list:
            cijizhan_no_list.append(0)

        self._measure_read()
        if not wait_for_finish:
            return

        time_start = time.time()
        unfinished_tag = copy.deepcopy(tag_no_list)

        while time.time() - time_start < timeout:
            self._measure_read()
            if wait_for_finish:
                ready_tag = []
                if len(unfinished_tag) == 0:
                    break
                for i in unfinished_tag:
                    nz_cnt = 0
                    for j in self.tag_list[i]['dist']:
                        if j != self.EMPTY_DIST_VAL:
                            nz_cnt += 1
                    if nz_cnt >= len(cijizhan_no_list):# 因为主基站也会测距，所以此处+1 TODO: 还要其他函数需要修改通样的？
                        ready_tag.append(i)
                    else:
                        logger.debug("nz_cnt:%d < %d",nz_cnt,len(cijizhan_no_list))

                for i in ready_tag:
                    unfinished_tag.remove(i)
                ready_tag.clear()

        if wait_for_finish:
            if len(unfinished_tag) > 0:
                logger.debug('测量标签位置失败:%s', unfinished_tag)
            else:
                logger.debug('测量标签位置完成:%s', tag_no_list)

    def _measure_all_tag_dist(self, tag_client_list, anchor_client_list=None, ready_cnt=30):
        """

        :param tag_client_list:[{client_id:,}, ...]
        :param anchor_client_list:[{client_id:,}, ...]
        :return:[{'1-1-0:'1.1, '1-1-1':3.4},...] every item is a dict for every tag dist with all anchors.
        """
        tag_client_id_list = [i['client_id'] for i in tag_client_list]
        tag_no_list = [client_id_get_no(i) for i in tag_client_id_list]

        if anchor_client_list is None:
            anchor_client_list, tinfo = self._get_anchor_and_tag_info()

        self.set_to_start(tag_no_list,ready_cnt)
        self.locate_loop(tag_no_list, 1, ready_cnt*(len(tag_client_list)+1)/10.0)

        self.stop_measure_h()
        self.tag_list_cache_ready_cnt = 10

        for tt in self.all_device_list:
            tag_no = client_id_get_no(tt['client_id'])
            if tag_no in tag_no_list:
                tt['dist'] = dist_list2dict(self.tag_list[tag_no]['dist'])

        dist_all = []
        for i in tag_client_list:
            dist_all.append(dist_list2dict(self.tag_list[client_id_get_no(i['client_id'])]['dist'], self.group_id))

        return dist_all

    def _start_anchor_dist_measure(self, client_id, ready_cnt=40):
        self.curr_anchor_to_measure = client_id  # FIXME: client_id[2:]
        self.bulk_mode = 0
        logger.info("基站自动定位测距：%s", client_id)

        self.bulk_mode = 0
        # 基站转标签
        is_ok, ret = self.convert_anchor_to_tag_h(client_id)
        if not is_ok:
            return is_ok
        time.sleep(0.25)

        # 设置基站和标签参数
        except_list = [client_id_get_no(client_id)]
        except_list = list(set(except_list))
        self.set_basic_info_h([0], except_list)

        # 启动测量
        self.tag_list_cache_ready_cnt = ready_cnt
        self.start_measure_h(6)
        return True

    def _tag_a_process(self, tag_id, acce_x, acce_y, acce_z):
        """向tag_a_xyz中添加 标签20 21 的加速度"""
        if tag_id not in self.tag_a_xyz_cache_count:
            self.tag_a_xyz_cache_count[tag_id] = 0
        self.tag_a_xyz_cache_count[tag_id] += 1
        if acce_x != '':
            self.tag_a_xyz_cache[tag_id]['x'].append(short(int.from_bytes(acce_x, 'big')) * 16 / 32768)
        if acce_y != '':
            self.tag_a_xyz_cache[tag_id]['y'].append(short(int.from_bytes(acce_y, 'big')) * 16 / 32768)
        if acce_z != '':
            self.tag_a_xyz_cache[tag_id]['z'].append(short(int.from_bytes(acce_z, 'big')) * 16 / 32768)
        if self.tag_a_xyz_cache_count[tag_id] > self.tag_list_cache_ready_cnt:
            # print('self.tag_a_xyz_cache: ', self.tag_a_xyz_cache)
            x_sum = 0
            for j1 in range(len(self.tag_a_xyz_cache[tag_id]['x'])):
                x_sum += self.tag_a_xyz_cache[tag_id]['x'][j1]
            if len(self.tag_a_xyz_cache[tag_id]['x']) > 0:
                x_avg = x_sum / len(self.tag_a_xyz_cache[tag_id]['x'])
                self.tag_a_xyz[tag_id]['x'] = x_avg
            y_sum = 0
            for j1 in range(len(self.tag_a_xyz_cache[tag_id]['y'])):
                y_sum += self.tag_a_xyz_cache[tag_id]['y'][j1]
            if len(self.tag_a_xyz_cache[tag_id]['y']) > 0:
                y_avg = y_sum / len(self.tag_a_xyz_cache[tag_id]['y'])
                self.tag_a_xyz[tag_id]['y'] = y_avg
            z_sum = 0
            for j1 in range(len(self.tag_a_xyz_cache[tag_id]['z'])):
                z_sum += self.tag_a_xyz_cache[tag_id]['z'][j1]
            if len(self.tag_a_xyz_cache[tag_id]['z']) > 0:
                z_avg = z_sum / len(self.tag_a_xyz_cache[tag_id]['z'])
                self.tag_a_xyz[tag_id]['z'] = z_avg

            self.tag_a_xyz_cache_count[tag_id] = 0
            self.tag_a_xyz_cache[tag_id] = {'x': [], 'y': [], 'z': []}

    def _decode_measure_payload(self, payload63: bytes, pkt_has_height=1):
        """
        解码报文
        :param payload63:
        :param pkt_has_height:
        :return:
        """
        tag_id, dista, battery_tag, battery_anchor, acce_x, acce_y, acce_z, status, noise_param = self.decode_measure(
            payload63, pkt_has_height)
        for i in range(len(self.battery_anchor)):
            if battery_anchor[i] == 255:
                continue
            self.battery_anchor[i] = battery_anchor[i]

        for i in list(battery_tag.keys()):
            if battery_tag[i] == 255:
                del battery_tag[i]
        self.battery_tag.update(battery_tag)

        self._tag_a_process(tag_id, acce_x, acce_y, acce_z)
        self._tag_dist_process(tag_id, dista, status)
        self._count_for_payload_func(tag_id, dista, status, acce_x, acce_y, acce_z)
        return

    def _count_for_payload_func(self, tag_id, dista, status, acce_x, acce_y, acce_z):
        """
        测试 运行速度
        :return:
        """

        for i in range(1, 8):
            # 错误统计
            tag_en = ((status & (1 << i)) >> i)
            tag_dist = (dista[i] / 100.0)
            if tag_en and (tag_dist != 0.0):
                self.biaoqian_jizhan_error[tag_id][i - 1][0] += 1
            self.biaoqian_jizhan_error[tag_id][i - 1][1] += 1

        self.count_for_payload += 1
        ctime = time.time()
        if ctime - self.count_for_payload_timer > 3.0:
            self.count_for_payload_pps = round(self.count_for_payload / (ctime - self.count_for_payload_timer), 1)
            logger.debug('count_for_payload: %s', self.count_for_payload_pps)
            logger.debug('error_status: %s', self.biaoqian_jizhan_error)
            self.count_for_payload = 0
            self.count_for_payload_timer = ctime

    def _avg_dist_data(self, data_list, last_data):
        """
            距离数据的平均滤波。避免跳动太大
        :param data_list:
        :param last_data:
        :return:
        """
        logger.debug('avg_dist_data 输入 %s %s', data_list, last_data)
        if len(data_list) == 0:
            logger.debug('avg_dist_data result:', last_data)
            return last_data, 0
        elif len(data_list) == 1 and last_data == self.EMPTY_DIST_VAL:
            # 针对第一个测距数据可能不对的问题，如果之前没有测距值，现在只有1个测距值，则不使用，等待有多个测距值再用
            return last_data, 0

        if len(data_list) > 5:
            data_list.sort()
            data_list = data_list[2:-2]
        elif len(data_list) > 3:
            data_list.sort()
            data_list = data_list[1:-1]

        dlen = len(data_list)
        dist_avg = sum(data_list) / len(data_list)
        logger.debug('avg_dist_data result: %s', dist_avg)
        if (math.fabs(dist_avg - last_data) > 1) and (last_data != self.EMPTY_DIST_VAL):
            logger.debug('距离跳跃太大：%s %s %s', data_list, dist_avg, last_data)
        return dist_avg, 1

    def _tag_dist_cache_to_one(self, tag_id):
        """
        标签的距离，采用多次测量取平均值的方式
        :return:
        """
        currdist_list = self.tag_list_cache[tag_id]['dist']
        logger.debug('tag_id= %s 距离测距次数:%s', tag_id, self.dist_decode_cnt[tag_id])
        logger.debug('距离测距数值:%s', currdist_list)

        for i in range(len(currdist_list)):
            anchor_index = i
            anchor_dist = currdist_list[i]

            ###统计基站测距次数和成功次数

            ###

            if len(anchor_dist) > 0:
                curr_val, curr_val_en = self._avg_dist_data(anchor_dist, self.tag_list[tag_id]['dist'][i])
                if tag_id == 0:
                    # 某一个次基站（0标签）正在测距
                    tag_id_c = self.curr_anchor_to_measure  # 1-2-0
                else:
                    # 某一个标签正在测距 20、21、1、2、3、4
                    tag_id_c = pack_client_id(self.group_id, DEVICE_TYPE_TAG, tag_id)
                #  tag_id：2-4或3-20，anchor_id：0-5之间的一个数字，curr_val：距离，cali_param：计算参数
                curr_val = self._data_cali_v2(client_id=tag_id_c, anchor_id_n=i, curr_val=curr_val,
                                              cali_param=self.cali_param)
                if i > 5:
                    break

                self.tag_list[tag_id]['dist'][i] = curr_val
                self.tag_list[tag_id]['dist_en'][i] = curr_val_en
                self.tag_list[tag_id]['dist_cache'][i] = currdist_list[i]

            self.tag_list_cache[tag_id]['dist'][i] = []

        self._one_tag_dist_got(tag_id)

    def _one_tag_dist_got(self, tag_id):
        """
        每次接收到1次设备发来的标签距离信息，则会调用此函数
        :param tag_id:
        :return:
        """

        ainfo, tinfo = self._get_anchor_and_tag_info()
        for index, i in enumerate(ainfo.copy()):
            if client_id_get_no(i['client_id'])==ANCHORZ_NO:
                #主基站
                del ainfo[index]
                break

        dist = dist_list2dict(self.tag_list[tag_id]['dist'], self.group_id)
        if tag_id != 0:
            pos = self.tag_locate_algorithm.calc(dist, pack_client_id(self.group_id, DEVICE_TYPE_TAG, tag_id), ainfo)
        else:
            pos = {'x': 0, 'y': 0, 'z': 0}
        if self.pos_callback is not None:
            self.pos_callback(tag_id, dist, pos)

    def _tag_dist_process(self, tag_id, dista, status):
        """
        处理距离
        :param dista:
        :param status:
        :param tag_id:
        :return:
        """
        if tag_id not in self.tag_list:
            print('tag_id not in self.tag_list:', tag_id, self.tag_list)
            raise ValueError('tag_id不在tag_list中' + str(tag_id))

        """向tag_list中添加对应标签到各次基站的距离"""
        for i in range(0, 8):
            tag_en = ((status & (1 << i)) >> i)
            tag_dist = (dista[i] / 100.0)
            if tag_en and (tag_dist != 0.0):
                self.tag_list_cache[tag_id]['dist'][i].append(tag_dist)
        # logger.info('打印标签到各基站的距离列表多次： %s %s ', tag_id, self.tag_list_cache[tag_id]['dist'])

        if tag_id not in self.dist_decode_cnt:
            self.dist_decode_cnt[tag_id] = 0
        self.dist_decode_cnt[tag_id] += 1

        if self.dist_decode_cnt[tag_id] > self.tag_list_cache_ready_cnt:
            # 转实际值
            self._tag_dist_cache_to_one(tag_id)
            self.dist_decode_cnt[tag_id] = 0

    def _data_cali_v2(self, client_id, anchor_id_n, curr_val, cali_param):
        """

        """
        client_id2 = client_id_remove_group(client_id)
        anchor_id = str(anchor_id_n)
        if (client_id2 in cali_param) and (anchor_id in cali_param[client_id2]):
            curr_val *= cali_param[client_id2][anchor_id]
        logger.debug('data_cali: %s %s %s', client_id2, anchor_id_n, curr_val)
        logger.debug('输出: %s ', curr_val)
        return curr_val

    def _measure_read(self, pkt_has_height=1):
        """
        读取串口数据并解码，用于主基站主动发数据的模式
        :param pkt_has_height:
        :return: 此次处理的报文数，此次处理的字节数
        """
        if self.measure_mode != 3 and self.measure_mode != 4 and self.measure_mode != 6:
            return 0, 0
        # 只有3，4模式是主动发送，需要处理报文
        self.read_buf += self._read_data(0.1)
        #print('measure_read: ', len(self.read_buf), self.read_error_cnt, self.read_ok_cnt)
        #print('measure_read: ', self.read_error_stat)

        pkt_length = 63
        if pkt_has_height == 0:
            pkt_length = 63 - 16

        pkt_list, self.read_buf, read_error_cnt, ret_pkt_process, ret_byte_process = split_packet(self.read_buf,
                                                                                                  pkt_length,
                                                                                                  self.read_error_stat)
        for i in pkt_list:
            self._decode_measure_payload(i, pkt_has_height)

        return ret_pkt_process, ret_byte_process

    def _check_anchor_dist_measure_finish(self):
        """
        读取串口测距报文，处理后，查看哪些测距已经完成
        :return:is_finish, 100
        """
        self._measure_read()

        anchor_len = len(self.curr_used_anchor)
        if anchor_len <= 0:
            print('anchor_len: ', anchor_len, self.anchor_used_dict)
            return True, 100

        nz_cnt = 0
        for i in self.tag_list[0]['dist']:
            if i != self.EMPTY_DIST_VAL:
                nz_cnt += 1

        if nz_cnt >= anchor_len:
            print(':测距完成：', self.tag_list[0]['dist'], nz_cnt, anchor_len)
            return True, 100
        if 0 not in self.dist_decode_cnt:
            return False, 0
        return False, [len(i) for i in self.tag_list_cache[0]['dist']]

    def _stop_anchor_dist_measure(self, client_id):

        self.stop_measure_h()
        # 转换标签到基站
        ret = self.convert_tag_to_anchor_h(client_id)
        self.tag_list_cache_ready_cnt = 10
        dist_ret = copy.deepcopy(self.tag_list[0]['dist'])
        print('stop_jizhan_dist_measure: ', dist_ret)
        return dist_ret

    def _measure_one_anchor_dist(self, client_id, measure_ready_cnt=40, remove_error_device=True):
        """
        测量某个基站到其他基站的距离
        :param client_id:  当前要测量的次基站client_id       次基站的id：eg: 1-2-1
        :param measure_ready_cnt: 测量次数
        :param remove_error_device: False/True 如果测距失败，则从可用列表中移除此设备
        :return:
        """

        # 每个次基站，进行测距

        print('次基站' + client_id + '正在测距中')

        start_ok = self._start_anchor_dist_measure(client_id, measure_ready_cnt)
        if (not start_ok) and remove_error_device:
            print("次基站连接失败 ： 改为不启用，即删除该基站信息", client_id)
            for i in self.all_device_list.copy():
                if i['client_id'] == client_id:
                    self.all_device_list.remove(i)
                    break

        cnt = 0
        while start_ok:
            cnt += 1
            is_finish, percent = self._check_anchor_dist_measure_finish()
            if is_finish:
                break

            print("次基站测距中 client_id： percent----", client_id, percent)

            time.sleep(0.1)
            if cnt > 100:  # 10秒种
                print("基站测距超时失败 ", client_id)
                time.sleep(0.1)
                break

        dist = self._stop_anchor_dist_measure(client_id)

        # dist的list->dict
        # 基站：client_id 到其它基站的距离在dist中，需要写入self.all_device_list
        for i in self.all_device_list:
            if i['client_id'] == client_id:
                i['dist'] = dist_list2dict(dist, self.group_id)
                break
        print('次基站' + client_id + '测距成功, dist:', dist)
        return dist_list2dict(dist, self.group_id)



import traceback
import copy
import time
from .calc_crc import check_crc
from .client_id_utils import client_id_get_no
from .modbus_pkt import get_read_pkt, get_write_pkt, decode_read_return, decode_write_return
import serial
from .mylog import logging
#from .detect_com_port import is_anchor0

import serial.tools.list_ports


logger = logging.getLogger(__name__)

#主基站编号，固定为0
ANCHORZ_NO=0

# old reg addr
REG_TAG_NUM = 0x29
REG_TAG_LIST = 0x37
REG_BAUD_RATE = 0x00
REG_MODBUS_ID = 0x01
REG_DEVICE_MODE = 3
REG_DEVICE_ID = 4
REG_CHANNEL_DATARAT = 5
REG_ANT_DLY = 8
KALMAN_PARAM_ADDR = 6
REG_MEASURE_MODE_ADDR = 2
REG_CHANGE_ANC_ID = 0x8d
REG_CALCULATE_EN = 0x28
REG_CHANGE_FLAG=0x8e


#new reg addr
REG_TEST=0
REG_FIRMWARE_VERSION=1
REG_BAUD_RATE=2
REG_MODBUS_ID=3
REG_MEASURE_MODE_ADDR=4
REG_DEVICE_MODE=5
REG_DEVICE_ID=6
REG_CHANNEL_DATARAT=7
REG_ANT_DLY=8
REG_ANC_ENABLED=9
#Calculate_EN：
REG_CALCULATE_EN=10
REG_TAG_NUM=11
#可用标签列表(依次为C寄存器byte0、C寄存器byte1、D寄存器byte0、D寄存器byte1......)
REG_TAG_LIST=12
REG_TAG_ID_NOW=62
REG_CALCULATE_FLAG=63
REG_DIST_A=64#=//主基站测距
REG_DIST_B=65
REG_DIST_C=66
REG_DIST_D=67
REG_DIST_E=68
REG_DIST_F=69
REG_DIST_G=70
REG_DIST_H=71
REG_ACCX_GET=72
REG_ACCY_GET=73
REG_ACCZ_GET=74
REG_ACCX_OWN=75
REG_ACCY_OWN=76
REG_ACCZ_OWN=77
REG_CHANGE_ANC_ID=78
#1为将基站改为标签，2-标签改基站。 3-设置完成. 4-设置失败 5--等待中
REG_CHANGE_FLAG=79
REG_GROUP_ID=80
REG_IAP_RESET=100# //复位重启寄存器



class UwbModbus:
    def __init__(self):

        self.EMPTY_DIST_VAL = -1  # 空的距离位置值
        self.verbose = 0  # --0 不打印，1--打印调试信息
        self.demo_inst = None
        self.port = None
        self.port_name = None  # COM4
        self.bulk_mode = 0  # 是否直接发送
        self.is_demo_mode = 0  # 演示模式
        self.mbus_register = bytearray(106 * 2)
        self.tag_list_cache_ready_cnt = 10  # 用于缓存几个数据，就可以进行平均放入tag_list中的设置。默认10各数据。
        self.read_error_cnt = 0
        self.read_error_stat = {}
        self.read_ok_cnt = 0
        self.read_buf = b''  # 报文接收缓冲区
        self.count_for_payload = 0
        self.count_for_payload_timer = time.time()
        self.count_for_payload_pps = 0  # packet per second

        self.anchor_used_dict = {0: 1, 1: 1, 2: 1, 3: 1, 4: 1, 5: 1, 6: 1}  # 当前启动的基站列表，默认是6个
        self.curr_used_anchor_no = [] #当前使用的基站号列表
        self.curr_used_tag_no=[] #当前使用的标签号列表
        self.set_cache_data_to_default()

    def is_anchorz(self, port_name):
        logger.debug('is_anchorz:%s', port_name)
        time.sleep(0.1)
        is_ok = self.connect(port_name)
        if not is_ok:
            return False
        time.sleep(0.1)

        count = self.port.inWaiting()
        if count != 0:
            #正在测量？上次未关闭设备?
            self.start_measure_h(0)
            time.sleep(0.5)
            count = self.port.inWaiting()
            self.port.read(count)
            count = self.port.inWaiting()
            self.port.read(count)

        is_ok, ret = self.modbus_test_read()
        if not is_ok:
            return False
        is_ok, ilist,bdat = self.read_modbus_h(REG_DEVICE_MODE,1)
        if not is_ok:
            return False

        device_mode = ilist[0]
        logger.debug('device mode:%s(%d)\n',self.device_model_to_str(device_mode), device_mode)

        self.disconnect()
        return True

    def device_model_to_str(self, mode_int):
        #设备模式 0：标签 1：次基站 2：主基站
        if mode_int==0:
            return 'tag'
        elif mode_int==1:
            return 'anchor'
        elif mode_int==2:
            return 'anchorz'

    def detect_com_port(self):
        """
        检测串口：
        :return: 'COM3'  对应uwb的主基站串口
                没有相应设备时，为None
        """
        ret = None
        port_list = list(serial.tools.list_ports.comports())
        if len(port_list) <= 0:
            # print("The Serial port can't find!")
            logger.warning("have no  Serial port find!")
            return ret

        for i in list(port_list):
            port_name = i[0]
            if self.is_anchorz(port_name):
                ret = port_name
                logger.debug('port find')

        return ret

    def disconnect(self):
        # 关闭串口
        if self.is_port_open():
            self.port.close()
            logger.debug('port disconnect')

    def modbus_test_write(self, dat=0):
        self.write_modbus_h(REG_TEST, 1,dat.to_bytes(2, byteorder='big', signed=False))

    def modbus_test_read(self):
        is_ok, ilist, bdat= self.read_modbus_h(REG_TEST, 1)
        if not is_ok:
            return False, None

        return is_ok, ilist[0]


    def connect(self, com_port=None):
        """连接串口"""
        try:
            self.port_name = com_port
            if self.port_name is None:
                self.port_name = self.detect_com_port()

            if self.port_name is None:
                print('could not find device')
                return False

            #print('port: ', self.port_name)
            #print('self.port_name：', self.port_name)

            self.port = serial.Serial(port=self.port_name, baudrate=115200, timeout=0.1)  # , timeout=0.1
            #print('self.port：', self.port)
            self.port.close()
            if not self.is_port_open():
                self.port.open()
            return True
        except Exception as e:
            print('connect failed', e)
            traceback.print_exc()
            return False

    def _demo_write_data(self, data):
        if self.demo_inst is None:
            return
        return self.demo_inst.write(data)

    def _write_data(self, data):
        logger.debug('write_data: %s', data.hex())

        if self.is_demo_mode:
            self._demo_write_data(data)
            return

        if self.port is None or (not self.is_port_open()):
            pass
            print('self.port is None or (not self.is_port_open())')
            # logger.error('send error. not open')
        else:
            try:
                self.port.write(data)
            except Exception as e:
                logger.error('串口错误? %s', e)
                self.port.close()
                self.port.open()
                self.port.write(data)

    def _demo_read_data(self, timeout, byte_max):
        if self.demo_inst is None:
            return
        # time.sleep(timeout)
        return self.demo_inst.read(timeout, byte_max)

    def _read_data(self, timeout=0.001, byte_max=-1):
        """

        :param timeout:
        :return:
        """

        data_get = b''
        t1 = time.time()
        while time.time() - t1 < timeout:
            try:
                if self.is_demo_mode:
                    data_get += self._demo_read_data(timeout, byte_max)
                else:
                    data_get += self.port.read(1024)
            except Exception as e:
                if self.is_demo_mode:
                    pass
                else:
                    logger.error('串口错误? %s', e)
                    self.port.close()
                    self.port.open()
                    data_get += self.port.read(1024)

            # logger.debug('read_data: %s', data_get.hex())
            if (byte_max > 0) and (byte_max <= len(data_get)):
                break

        logger.debug('read_data_ret: %s', data_get.hex())

        return data_get

    def check_crc(self, payload: bytes):
        """

        :param payload: 带校验和的报文，最后2byte的校验和会被用于比较
        :return: 正确返回True, 否则返回false
        """
        return check_crc(payload)

    def read_modbus_h(self, reg_start_addr, reg_count, addr=1):
        """

        :param reg_start_addr:
        :param reg_count:
        :param addr:
        :return: is_ok, ilist, bytes_data
        """
        """
            指定id、地址、寄存器个数
            01 --id号： id
            03-- 读
            00 05 --起始寄存器  add2
            00 6A -- 寄存器个数  reg_count
            C5 E5 --校验位

            01 --id号： id
            03 -- 读
            02 -- 字节数
            0007 -- 数据
            f986
            寄存器是16bit的。低位在前，高位在后。
        """

        to_w = get_read_pkt(reg_start_addr, reg_count, addr)

        self._write_data(to_w)
        to_read_len = 5 + reg_count * 2
        ret = self._read_data(1, byte_max=to_read_len)

        ret2 = decode_read_return(ret, reg_count)
        is_ok=False
        dat = None
        if 'crc_ok' in ret2:
            is_ok = ret2['crc_ok']

        if is_ok:
            ilist =[]
            ll = len(ret2['data'])//2
            for i in range(ll):
                dat = int.from_bytes(ret2['data'][i*2:i*2+2], 'big', signed=False)
                ilist.append(dat)
        else:
            ilist = []
            ret2['data'] = ret

        return is_ok,ilist, ret2['data']


    def _write_modbus_real(self, add2, reg_count, dat, id=1, just_gen_pkt=0):
        """ 例子：开始测距 01 10 00 28 00 01 02 00 04 A1 BB
                    01 id号  id
                    10 写
                    00 00 起始寄存器  add2
                    00 6A 寄存器个数  reg_count
                    D4   字节数      len(dat)
                    。。。。 D4个字节  dat
                    0B 60 --校验和

                    返回值
                    00
                    10
                    006b   107
                    0001   1
                    71c4

                    01  id号  id
                    10  写
                    006b  起始寄存器
                    0001  寄存器个数
                    7015  校验位


                :param id:
                :param add2:
                :param reg_count:
                :param dat:
                :return:
                """
        logger.debug('write_modbus_real: %s %s %s %d', add2.to_bytes(2, 'big').hex(), reg_count.to_bytes(1, 'big').hex()
                     ,
                     dat.hex(), id)

        to_r = get_write_pkt(add2, reg_count, dat, id)

        if just_gen_pkt:
            # logger.debug("write_modbus_real just_gen_pkt=1 do not send")
            return to_r

        self._write_data(to_r)
        ret = self._read_data(1, 8)

        ret2= decode_write_return(ret)
        #print(ret2)
        return ret2

    def write_modbus_h(self, add2, reg_count, dat, id=1, just_gen_pkt=0):
        """ 例子：开始测距 01 10 00 28 00 01 02 00 04 A1 BB
            01 id号  id
            10 写
            00 00 起始寄存器  add2
            00 6A 寄存器个数  reg_count
            D4   字节数      len(dat)
            。。。。 D4个字节  dat
            0B 60 --校验和

            返回值
            00
            10
            006b   107
            0001   1
            71c4

            01  id号  id
            10  写
            006b  起始寄存器
            0001  寄存器个数
            7015  校验位


        :param id:
        :param add2:
        :param reg_count:
        :param dat:
        :return:
        """
        logger.debug('write_modbus: %s %s %s %d', add2.to_bytes(2, 'big').hex(), reg_count.to_bytes(1, 'big').hex(),
                    dat.hex(), id)

        if self.bulk_mode and (not just_gen_pkt):
            # logger.debug('write in bulk mode. just write to local without return data')
            # bulk mode，just write to buffer
            for i in range(reg_count * 2):
                self.mbus_register[add2 * 2 + i] = dat[i]

            return {'raw_data': b''}

        return self._write_modbus_real(add2, reg_count, dat, id, just_gen_pkt=just_gen_pkt)

    def bulk_send(self):
        # logger.debug('bulk_send')
        self._write_modbus_real(0, len(self.mbus_register) // 2, self.mbus_register)


    def read_label_h(self):
        is_ok, ret, bdata = self.read_modbus_h(REG_TAG_NUM, 51)
        if is_ok:
            label_num = ret[0]
            return is_ok, ret[1:label_num+1]
        return is_ok, ret

    def set_label_h(self, label_no_list=(0,)):
        """
        如果label_no_list=[1] 则发送数据为 00 01
        如果label_no_list=[1,2] 则发送数据为 02 01
        如果label_no_list=[1,2,3] 则发送数据为 02 01 00 03
        :param label_no_list:
        :return:
        """
        # logger.info('设置标签列表：%s', json.dumps(label_no_list))
        label_no_list = copy.deepcopy(label_no_list)
        label_cnt = len(label_no_list)

        if label_cnt % 2 == 1:
            # 标签是奇数个，则+0变为偶数个
            label_no_list.append(0)

        label_dat = b''
        for i in range(len(label_no_list) // 2):
            label_dat += label_no_list[i * 2 + 1].to_bytes(1, 'big') + label_no_list[i * 2].to_bytes(1, 'big')

        self.write_modbus_h(REG_TAG_LIST, (label_cnt + 1) // 2, label_dat)
        # 因为设置标签都是bulk模式，所以此处不需要sleep
        if not self.bulk_mode:
            time.sleep(0.5)

    def read_label_num_h(self):
        is_ok, data, bdata = self.read_modbus_h(REG_TAG_NUM, 1)
        return is_ok, data
    def set_label_num_h(self, num: int):
        # logger.info('设置标签数：%d', num)
        num_b = num.to_bytes(2, 'big')
        self.write_modbus_h(REG_TAG_NUM, 1, num_b)
        # 因为此处是bulk模式，所以此处不需要sleep
        if not self.bulk_mode:
            time.sleep(0.5)

    def set_buardrate_h(self, rate_no=7):
        """
        设备串口通讯波特率
        :param rate_no:0：4800  1：9600 2：14400 3：19200 4：38400 5：56000 6：57600 7：115200  8：128000 9：256000
        :return:
        """
        self.write_modbus_h(REG_BAUD_RATE, 1, rate_no.to_bytes(2, 'big'))

    def set_modbus_id_h(self, id=1):
        self.write_modbus_h(REG_MODBUS_ID, 1, id.to_bytes(2, 'big'))

    def read_modbus_id_h(self):
        is_ok, ret, bdata = self.read_modbus_h(REG_MODBUS_ID,1, 0xff)
        return is_ok, ret

    def set_device_measure_mode_h(self, cmode=1, dmode=0):
        """
        :param cmode:测量模式   0：DS-TWR 1：高性能TWR。
        :param dmode:测距模式 1:二维模式 2：三维模式"
        :return:
        """
        self.write_modbus_h(REG_MEASURE_MODE_ADDR, 1, cmode.to_bytes(1, 'big') + dmode.to_bytes(1, 'big'))

    def set_device_mode_h(self, mode=2):
        """

        :param mode:设备模式 0：tag 1：anchor 2：anchorz
        :return:
        """
        if mode=='tag':
            mode = 0
        elif mode=='anchor':
            mode = 1
        elif mode=='anchorz':
            mode = 2

        if mode<0 or mode>2:
            logger.error('mode not known %s', mode)

        self.write_modbus_h(REG_DEVICE_MODE, 1, mode.to_bytes(2, 'big'))
    def read_device_mode_h(self):
        is_ok, ilist, bdat = self.read_modbus_h(REG_DEVICE_MODE, 1)
        if not is_ok:
            return False, None

        return is_ok, ilist[0]

    def set_modbus_id_h(self, mid):
        """

        :param mid: modbus_id/addr to set.
        :return:
        """
        self.write_modbus_h(REG_MODBUS_ID, 1, mid.to_bytes(2, 'big'), id=0xff)


    def set_device_id_h(self, device_id=0):
        """
        设备ID，高8位为次基站ID，范围0~6 ，
        低8位为标签ID ，0~99
        （程序内部 标签ID为0~247  次基站ID为248~254  主基站ID为255）
        :param device_id:
        :return:
        """
        self.write_modbus_h(REG_DEVICE_ID, 1, device_id.to_bytes(2, 'big'))

    def read_device_id_h(self):
        is_ok, ilist, bdat = self.read_modbus_h(REG_DEVICE_ID, 1)
        if not is_ok:
            return False, None

        return is_ok, ilist[0]

    def read_comm_channel_and_speed_h(self):
        is_ok, ret, bdata = self.read_modbus_h(REG_CHANNEL_DATARAT, 1)
        return is_ok, ret
    def set_comm_channel_and_speed_h(self, channel=1, speed=2):
        """
        byte0-空中信道，byte1-空中传输速率

        :param channel:
        :param speed:
        :return:
        """
        self.write_modbus_h(REG_CHANNEL_DATARAT, 1, channel.to_bytes(1, 'big') + speed.to_bytes(1, 'big'))

    def set_kalman_h(self, param_q=3, param_r=0x0a):
        """

        :param param_q:
        :param param_r:
        :return:
        """
        self.write_modbus_h(KALMAN_PARAM_ADDR, 2, param_q.to_bytes(2, 'big') + param_r.to_bytes(2, 'big'))

    def set_recv_delay_h(self, delay=0x80CF,
                         just_gen_pkt=0):  # 天线延迟  ===================================================================
        """

        :param delay:
        :return:
        """
        return self.write_modbus_h(REG_ANT_DLY, 1, delay.to_bytes(2, 'big'), just_gen_pkt=just_gen_pkt)

    def read_anchor_enable_h(self):
        is_ok , ret, bdata = self.read_modbus_h(REG_ANC_ENABLED, 1, 1)
        ret_anchor_enabled=[]
        if is_ok:
            for i in range(1,7):
                if(ret[0]&(1<<i)):
                    ret_anchor_enabled.append(i-1)
        return is_ok, ret_anchor_enabled

    def set_anchor_enable_h(self,except_anchor_list=None):
        enable_byte=0xff
        for i in range(8):
            if except_anchor_list is not None:
                if i in except_anchor_list:
                    enable_byte &= ~(1<<(i))
        logger.debug('set anchor enable:%08x', enable_byte)
        self.write_modbus_h(REG_ANC_ENABLED, 1, enable_byte.to_bytes(2, 'big'))


    def set_anchor_enable_old_h(self, jizhan_index=0, enabled=1):
        """
        基站固定的有7个。编号为0-6. 对应 BCDEFGH
        :param jizhan_index: 基站编号 0-6
        :param enabled: 0--disable. 1--enable it
        :return:
        """
        """
        基站固定的有7个。编号为0-6. 对应 BCDEFGH
        :param jizhan_index:基站编号 0-6
        :return:
        """
        if jizhan_index > 6 or jizhan_index < 0:
            # logger.error("基站编号超过范围：%s", jizhan_index)
            return
        # logger.info('设置基站使能 %d %d', jizhan_index, enabled)
        # FIXME: 此处基站使能地址不连续???

        jizhan_enable_addr = [12, 16, 20, 24, 28, 32, 36]
        self.write_modbus_h(jizhan_enable_addr[jizhan_index], 1, enabled.to_bytes(2, 'big'))

    def start_measure_h(self, mode=6):
        """
        :param mode:   0x04：持续测量，主动发送，不写入寄存器
                    0x03：单次测量，主动发送，写入寄存器
                    0x02：持续测量，不发送，写入寄存器
                    0x01：单次测量，不发送，写入寄存器
                    0x00: 停止测量"
        :return:
        """
        # logger.info('start_measure %d', mode)
        self.measure_mode = mode
        self.write_modbus_h(REG_CALCULATE_EN, 1, mode.to_bytes(2, 'big'))
        time.sleep(0.3)

    def stop_measure_h(self):
        # 停止连续测量
        # logger.info('停止测量:')
        self.start_measure_h(0)
        data = self._read_data(1, -1)
        # logger.debug('unused pkt: %s', data.hex())

    def is_port_open(self):
        if self.is_demo_mode:
            if self.demo_inst is None:
                return False
            if self.demo_inst.is_connected():
                return True
            return False

        if self.port is None or (not self.port.isOpen()):
            return False
        return True

    def is_convert_complete_h(self):
        is_ok, ilist, bdat = self.read_modbus_h(REG_CHANGE_FLAG, 1)
        if  not is_ok:
            return False
        if ilist[0] != 0x03:
            logger.warning('is_convert_complete: %s', bdat.hex())
            return False
        logger.debug('is_convert_complete OK')
        return True

    def convert_tag_to_anchor_once_h(self, client_id):
        """
        将标签 转化为 次基站
        """
        print('将标签 转化为 次基站: 次基站 ', client_id)
        # 将标签0改为B基站 ：  01 10 00 8d 00 02 04 00 00 00 02 A1 BB
        # 开始修改 ：  01 10 00 28 00 01 02 00 05 A1 BB
        jizhan_id = int(client_id_get_no(client_id))

        #FIXME: 此处基站id=1-8. 单片机中此处参数是0-7.所以修改-1
        jizhan_id=jizhan_id-1
        jizhan_id=jizhan_id.to_bytes(1, 'big')
        self.write_modbus_h(REG_CHANGE_ANC_ID, 2, jizhan_id + b'\x00\x00\x02')
        time.sleep(0.5)
        ret = self.write_modbus_h(REG_CALCULATE_EN, 1, b'\x00\x05')
        time.sleep(0.5)
        return ret

    def convert_tag_to_anchor_h(self, client_id, loop_max_cnt=5):
        ret = {}
        for j in range(loop_max_cnt):
            ret = self.convert_tag_to_anchor_once_h(client_id)
            if not self.is_demo_mode:
                time.sleep(0.3)
            is_ok = self.is_convert_complete_h()
            if is_ok:
                # logger.debug('convert_biaoqian_to_jizhan ok 2 %s', client_id)
                return is_ok, ret

            if not self.is_demo_mode:
                time.sleep(1)

        # logger.error("convert_biaoqian_to_jizhan %s", client_id)
        return False, ret

    def convert_anchor_to_tag_once_h(self, client_id):
        """
        将次基站 转化为 标签0
        """
        # logger.info('将次基站 转化为 标签: 次基站 %s', client_id)
        #                   01 10 00 8d 00 02 04 00 00 00 01 00 00
        # 修改B基站为标签0 ：  01 10 00 8d 00 02 04 00 00 00 01 A1 BB
        #            01 10 00 28 00 01 02 00 05 00 00
        # 开始修改 ：  01 10 00 28 00 01 02 00 05 A1 BB

        jizhan_id = int(client_id_get_no(client_id) )
        #FIXME: 此处基站id=1-8. 单片机中此处参数是0-7.所以修改-1
        jizhan_id=jizhan_id-1
        jizhan_id = jizhan_id.to_bytes(1, 'big')
        self.write_modbus_h(REG_CHANGE_ANC_ID, 2, jizhan_id + b'\x00\x00\x01')
        time.sleep(0.2)
        ret = self.write_modbus_h(REG_CALCULATE_EN, 1,b'\x00\x05')
        return ret

    def convert_anchor_to_tag_h(self, client_id, loop_max_cnt=5):
        self.bulk_mode = 0
        ret = None
        for i in range(loop_max_cnt):
            ret = self.convert_anchor_to_tag_once_h(client_id)
            if not self.is_demo_mode:
                time.sleep(0.2)
            is_ok = self.is_convert_complete_h()
            if is_ok:
                # logger.debug('convert_jizhan_to_biaoqian:OK-OK')
                return is_ok, ret

            # logger.warning('convert_jizhan_to_biaoqian again:%s', client_id)
            if not self.is_demo_mode:
                time.sleep(0.1)
        # logger.error('convert_jizhan_to_biaoqian %s', client_id)
        return False, ret

    def get_dist_of_tag_once_inner_h(self):
        """
            获取某个标签的距离:
            测量3次，取平均值
        """
        # logger.debug('get_dist_of_biaoqian_once: start')
        self.start_measure_h(1)
        time.sleep(1)
        # FIXME: 此处需要确认地址是从哪个基站开始，改为从主基站开始???
        is_ok, ilist, bdat = self.read_modbus_h(REG_DIST_A, 8)

        dist_b = [self.EMPTY_DIST_VAL for i in range(8)]
        if is_ok:
            for i in range(8):
                dist_b[i] = ilist[i]/100.0

        self.stop_measure_h()
        # time.sleep(0.5)
        print('标签到各基站距离: %s', dist_b)
        return 0, 0, dist_b

    def set_cache_data_to_default(self):
        """标签20 21 到各次基站的距离"""
        # 每个标签的距离
        v = self.EMPTY_DIST_VAL
        self.tag_list_item_default={'dist': [v, v, v, v, v, v, v, v], 'dist_en': [0, 0, 0, 0, 0, 0, 0, 0],
                              'dist_cache': [[], [], [], [], [], [], [], []]}
        self.tag_list = {0:copy.deepcopy(self.tag_list_item_default),
                         }
        for i in self.curr_used_tag_no:
            self.tag_list[i] = copy.deepcopy(self.tag_list_item_default)

        self.tag_list_cache_item_default={'dist': [[], [], [], [], [], [], [], []]}
        # 每个标签的距离缓存，每秒钟会有总共约48个距离，取平均后放入tag_list
        self.tag_list_cache = {1: copy.deepcopy(self.tag_list_cache_item_default),
                               }
        for i in self.curr_used_tag_no:
            self.tag_list_cache[i] = copy.deepcopy(self.tag_list_cache_item_default)


        self.tag_list_cache_ready_cnt = 10  # 用于缓存几个数据，就可以进行平均放入tag_list中的设置。默认10各数据。

        self.dist_decode_cnt = {1: 0}

        """标签20 21 的加速度"""
        # 标签20 21 加速度
        self.tag_a_xyz_item_default = {'x': 0, 'y': 0, 'z': 1}
        self.tag_a_xyz = {1: copy.deepcopy(self.tag_a_xyz_item_default),}
        for i in self.curr_used_tag_no:
            self.tag_a_xyz[i] = copy.deepcopy(self.tag_a_xyz_item_default)



        # 标签20 21 缓存加速度，多个加速度求平均值后，添加到 tag_a_xyz
        self.tag_a_xyz_cache_item_default = {'x': [], 'y': [], 'z': []}

        self.tag_a_xyz_cache = {20: copy.deepcopy(self.tag_a_xyz_cache_item_default),                                }
        for i in self.curr_used_tag_no:
            self.tag_a_xyz_cache[i] = copy.deepcopy(self.tag_a_xyz_cache_item_default)

        self.tag_a_xyz_cache_count = {}

        # 标签测距错误数、测距总次数
        self.biaoqian_jizhan_error = {20: [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0]],
                                      }
        for i in self.curr_used_tag_no:
            self.biaoqian_jizhan_error[i] = [[0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0], [0, 0]]


        # 标签位置
        self.tag_pos = {20: {'pos_x': 0, 'pos_x_max': 0, }}
        for i in self.curr_used_tag_no:
            self.tag_pos[i] = {'pos_x': 0, 'pos_x_max': 0, }


        # 标签位置的最大值，最小值
        self.tag_pos_error = {20: {'x': [0, 0], 'y': [0, 0], 'z': [0, 0]},
                              }
        for i in self.curr_used_tag_no:
            self.tag_pos_error[i] = {'x': [0, 0], 'y': [0, 0], 'z': [0, 0]}

    def clear_measure1_reg(self):
        """
        清除寄存器0x2f->0x36 8个寄存器
        :return:
        """
        ret = self.write_modbus_h(0x2f, 8, b'\x00' * 16)

    def set_basic_info_h(self, tag_no_list, except_anchor_list=None):
        """
        设置测距的基本信息。会通过bulk模式发送
        :param tag_no_list:[20,21,22,23,24] tag_no_list that will be used
        :param except_anchor_list: anchor that will be excluded. 默认的是空。
        :return:
        """

        self.bulk_mode = 1
        # 基站: B H :14379
        # 基本设置
        self.set_buardrate_h()
        self.set_modbus_id_h(1)
        self.set_device_measure_mode_h()
        self.set_device_mode_h()
        self.set_device_id_h()
        self.set_comm_channel_and_speed_h(2, 2)
        #self.set_kalman_h()
        self.set_recv_delay_h()
        #self.clear_measure1_reg()

        # 使能基站
        if except_anchor_list is None:
            except_anchor_list = []
        except_anchor_list.extend([i for i in self.anchor_used_dict if not self.anchor_used_dict[i]])

        self.curr_used_anchor_no = []
        for i in range(7):
            if except_anchor_list is not None:
                if i in except_anchor_list:
                    continue
            self.curr_used_anchor_no.append(i)

        self.set_anchor_enable_h(except_anchor_list)


        #self.curr_anchor_stat = {i: {'cnt': 0, 'success': 0} for i in self.curr_used_anchor}
        self.curr_used_tag_no=copy.deepcopy(tag_no_list)
        self.set_label_h(tag_no_list)  # 使用标签
        self.set_label_num_h(len(tag_no_list))  # 标签数量 1

        self.bulk_send()
        self.bulk_mode = 0

        # curr_used_tag_no curr_used_anchor_no updated up.
        self.set_cache_data_to_default()

        time.sleep(0.1)  # 0.5

    def decode_measure(self, payload63: bytes, pkt_has_height=1):
        battery_anchor = []
        battery_tag = {}
        tag_id = int.from_bytes(payload63[3:5], 'big')  # 标签id
        status_en = int.from_bytes(payload63[5:7], 'big')

        tag_xyz = payload63[7:13]
        dist_abcdefgh = payload63[13:29]
        max_noise = payload63[29:31]
        std_noise = payload63[31:33]
        path_param = payload63[33:45]
        if pkt_has_height:  # 新的 可以拿到 基站、标签电量信息
            acce_x = payload63[45:47]  # x重力加速度
            acce_y = payload63[47:49]
            acce_z = payload63[49:51]
            """
            20211221吊车实地测试问题： -----------z数据： 角度会有跳跃波动问题 ??? 有复现，已解决
                解决方法：当acce_x、acce_y、acce_z值均为b'\x00\x00'时，将角速度置为空
            """
            if acce_x == b'\x00\x00' and acce_y == b'\x00\x00' and acce_z == b'\x00\x00':
                acce_x = b''
                acce_y = b''
                acce_z = b''
            no_use_xyz = payload63[51:61]  # 电量： ABCDEFGH、当前测得标签的电量、FF占位
            crc = payload63[61:63]
        else:  # 原始代码
            acce_x = b''
            acce_y = b''
            acce_z = b''
            no_use_xyz = b''
            crc = payload63[45:47]

        # 处理 电量
        # DONE: 电量的显示
        battery_dict = {}
        if len(no_use_xyz) > 0:
            ba = []
            for i in range(8):
                battery_anchor.append(int.from_bytes(no_use_xyz[i:i + 1], 'big'))

            battery_tag[tag_id] = int.from_bytes(no_use_xyz[8:9], 'big')

        # 处理距离
        """
        dista = [0 for i in range(8)]
        dista[0] = int.from_bytes(dist_abcdefgh[0:2], 'big')
        dista[1] = int.from_bytes(dist_abcdefgh[2:4], 'big')
        dista[2] = int.from_bytes(dist_abcdefgh[4:6], 'big')
        dista[3] = int.from_bytes(dist_abcdefgh[6:8], 'big')
        dista[4] = int.from_bytes(dist_abcdefgh[8:10], 'big')
        dista[5] = int.from_bytes(dist_abcdefgh[10:12], 'big')
        dista[6] = int.from_bytes(dist_abcdefgh[12:14], 'big')
        dista[7] = int.from_bytes(dist_abcdefgh[14:16], 'big')
        """
        dista = []
        for i in range(8):
            dista.append(int.from_bytes(dist_abcdefgh[i * 2:i * 2 + 2], 'big'))
        logger.debug('tag_id=%s dista=%s en=%s battery_tag=%s battery_anchor=%s accel=%s %s %s',
                     tag_id, dista, status_en, battery_tag, battery_anchor, acce_x, acce_y, acce_z)
        return tag_id, dista, battery_tag, battery_anchor, acce_x, acce_y, acce_z, status_en, [max_noise, std_noise,
                                                                                               path_param]

    def is_anchor_exist_h(self, client_id, loop_max=3):
        self.bulk_mode = 0
        # logger.info('if_jizhan_exist: client_id=%s', client_id)
        is_ok, data_ret = self.convert_anchor_to_tag_h(client_id, loop_max)

        is_ok2, data_ret2 = self.convert_tag_to_anchor_h(client_id, loop_max)

        if is_ok or is_ok2:
            # logger.info('client_id=%s exists', client_id)
            return True
        else:
            # logger.info('client_id=%s not exists', client_id)

            return False

    def if_tag_exist_h(self, client_id, loop_max=5):
        client_no = client_id_get_no(client_id)
        self.set_basic_info_h([client_no], [1, 2, 3, 4, 5, 6])
        for i in range(loop_max):
            x, y, dist = self.get_dist_of_tag_once(1)
            for d in dist:
                if (d != 0) and (d != 0.0) and (d != self.EMPTY_DIST_VAL):
                    return True

        return False

    def get_dist_of_tag_once(self, loop_mean_cnt=20):
        """
        获取标签距离，默认平均20次
        """
        dist_all = []
        dist_list = []

        for i in range(loop_mean_cnt):
            x, y, dist_b = self.get_dist_of_tag_once_inner_h()
            dist_list.append(dist_b)
            # print(str(i)+'-次数---------: ', dist_b)
        # 求多次测距平均值
        d_dict = self.convert_mean(dist_list)
        # print('平均值: ', d_dict)
        return 0, 0, d_dict

    def convert_mean(self, dist_list):
        """
        求均值函数
        :param dist_list:
        :return:
        """
        d_dict = [[], [], [], [], [], [], [], []]

        # 将多次测量结果整合
        for i in dist_list:
            for j in range(7):
                d_dict[j].append(i[j])
        # 去掉值为0的数据
        for i in d_dict:
            new_l = []  # 存放不为0的数据
            for j in i:
                if j != 0:
                    new_l.append(j)
            i_index = d_dict.index(i)
            d_dict[i_index] = new_l
        for i in d_dict:
            i.sort()
            if len(i) >= 3:  # 列表长度大于等于3时，去掉最大最小值
                index = d_dict.index(i)
                d_dict[index] = i[1:-1]
        #  求距离 平均值
        for i in d_dict:
            if len(i) > 0:
                avg = sum(i) / len(i)
            else:
                avg = 0
            i_index = d_dict.index(i)
            d_dict[i_index] = avg

        return d_dict



















class AnchorLocateBase():
    def __init__(self):
        self.anchor_list=None

    def set_anchor_info(self, anchor_list):
        """
        :param anchor_list:[
        {
            client_id:'1-2-3',
            dist:{'1-2-1':10.5,... },
            direction_point:'N'
        },
        ...
          ]
        :return:
        """

        self.anchor_list = anchor_list

    def calc(self):
        """

        :return:
        """
        pass

    def get_anchor_pos(self):
        """

        :return:{
            client_id:'1-2-3',
            pos:{x:x, y:y, z:z}
        }
        """
        return self.anchor_list






from .dingwei import Dingwei

from .mylog import logging

logger = logging.getLogger(__name__)

class TagLocateAlgorithm1(Dingwei):
    def __init__(self):
        super().__init__()

    def calc(self, tag_dist, tag_client_id, anchor_pos=None, using_kalman= None):

        """
        传递基站位置信息和标签与基站的距离信息，返回标签位置
        :param anchor_pos:[{client_id:1-2-3, pos:{x:,y:,z:}}...]
        :param tag_dist: {1-2-3:10.4, 1-2-0:5,...}
        :return:{x:,y:， z:}
        """
        jizhan_pos_param = anchor_pos

        tag_pos_dict = {}

        # {'client_id': '1-3-1', 'dist': {'1-2-0': 5 * (10 ** 0.5), '1-2-1': 5 * (10 ** 0.5), '1-2-2': 5 * (10 ** 0.5), '1-2-3': 5 * (10 ** 0.5)，...}, ...}
        tag_dict = {'dist': tag_dist, 'client_id': tag_client_id}
        tag_pos_one_list = self.dingwei_biaoqian_dist_to_pos(jizhan_pos_param, [tag_dict], using_kalman=None)
        logger.debug('tag_pos_one_list %s', tag_pos_one_list)
        return tag_pos_one_list[0]['pos']



def test_demo1():
    anchor_pos = [
        {'client_id': '1-2-0', 'pos': {'x': 0, 'y': 30, 'z': 0.5}},
        {'client_id': '1-2-1', 'pos': {'x': 0, 'y': 0, 'z': 1.1}},
        {'client_id': '1-2-2', 'pos': {'x': 10, 'y': 0, 'z': 0.5}},
        {'client_id': '1-2-3', 'pos': {'x': 10, 'y': 30, 'z': 1.1}}
    ]
    tag_dist = {'1-2-0': 5 * (10 ** 0.5), '1-2-1': 5 * (10 ** 0.5), '1-2-2': 5 * (10 ** 0.5), '1-2-3': 5 * (10 ** 0.5)}
    t = TagLocateAlgorithm1()
    t.calc(tag_dist, '1-3-1', anchor_pos)


if __name__ == '__main__':
    test_demo1()



class KalmanFilterXYZ:
    def __init__(self,Q=3.0, R=10.0):
        self.kx=KalmanFilter(Q, R)
        self.ky=KalmanFilter(Q, R)
        self.kz=KalmanFilter(Q, R)
    def set_x(self,x):
        return self.kx.set_val(x)
    def set_y(self, y):
        return self.ky.set_val(y)
    def set_z(self, z):
        return self.kz.set_val(z)
class KalmanFilter:
    def __init__(self, Q=3.0, R=10.0):
        self.p_last = 0
        self.ProcessNiose_Q = Q
        self.MeasureNoise_R=R
        self.val_last = None
    def set_val(self, val):
        if self.val_last is None:
            self.val_last=val
        # 卡尔曼滤波 基于下一时刻位置不变模型
        p_mid = self.p_last + self.ProcessNiose_Q
        # 预测本次误差
        kg = p_mid / (p_mid + self.MeasureNoise_R)
        # 更新本次卡尔曼增益
        data_now = self.val_last + kg * (val - self.val_last)
        # 根据本次观测值预测本次输出
        p_now = (1 - kg) * p_mid
        # 更新误差
        self.p_last = p_now
        self.val_last = data_now
        return data_now


def Filter_Kalman(data_now, data_last, p_last, ProcessNiose_Q, MeasureNoise_R):
    """

    :param data_now: 数值 x/y/z
    :param data_last:  上一次的数值
    :param p_last:
    :param ProcessNiose_Q: 3
    :param MeasureNoise_R: 10
    :return:
    """
    # 卡尔曼滤波 基于下一时刻位置不变模型
    result = [[], []]
    p_mid = p_last + ProcessNiose_Q
    # 预测本次误差
    kg = p_mid / (p_mid + MeasureNoise_R)
    # 更新本次卡尔曼增益
    data_now = data_last + kg * (data_now - data_last)
    # 根据本次观测值预测本次输出
    p_now = (1 - kg) * p_mid
    # 更新误差
    p_last = p_now
    result[0] = data_now
    result[1] = p_last
    return result

def test_kf1():
    x=[1,2,1,2,1]
    p_last = 0
    x_dn=[]
    last_val = 1.0
    for i in range(len(x)):
        ret= Filter_Kalman(x[i],last_val, p_last, 3, 10)
        p_last = ret[1]
        print('ret:',i, ret[0])
        last_val = ret[0]
        x_dn.append(ret[0])

    assert x_dn[0]==1.0
    assert x_dn[1]==1.3467336683417086
    assert x_dn[2]==1.2105584375953615
    assert x_dn[3]==1.5336301851417857
    assert x_dn[4]==1.3122030348062983


def test_kf2():
    x=[1,2,1,2,1]
    kf2 = KalmanFilter()
    x_dn=[]
    for i in range(len(x)):
        curr_val = kf2.set_val(x[i])
        print('ret:',curr_val)
        x_dn.append(curr_val)

    assert x_dn[0]==1.0
    assert x_dn[1]==1.3467336683417086
    assert x_dn[2]==1.2105584375953615
    assert x_dn[3]==1.5336301851417857
    assert x_dn[4]==1.3122030348062983

if __name__=='__main__':
    test_kf1()
    test_kf2()





from .client_id_utils import client_id_get_no, pack_client_id, DEVICE_TYPE_ANCHORZ, DEVICE_TYPE_ANCHOR


def dist_dict2list():
    pass

def dist_list2dict(dist_list, group_id=1):
    """
    将 距离列表 转换成 距离字典
    :param dist_list:[1,2,3,4,5,6,7,8]
    :param group_id:3
    :return:{1-2-0:1, 1-2-1:2, 1-2-2:3,...}
    """
    if not isinstance(dist_list, list):
        # 如果dist_list 不是列表，则不作处理
        return dist_list
    dist_dict = {}
    for n in range(len(dist_list)):
        one_client_id = pack_client_id(group_id, DEVICE_TYPE_ANCHOR, n)
        dist_dict[one_client_id] = dist_list[n]
    return dist_dict


if __name__ == '__main__':
    ret = dist_list2dict([1, 2, -1, 4, 5, 6, 7, 8], 1)
    # ret = dist_list2dict('1-2-0', {}, 1)
    print(ret)




